#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, List

from tfx.proto import transform_pb2

from zenml.enums import StepTypes
from zenml.steps import BaseStep
from zenml.steps.trainer.utils import TRAIN_SPLITS

SPLIT_MAPPING = 'split_mapping'


def build_split_mapping(args):
    if SPLIT_MAPPING in args and args[SPLIT_MAPPING]:
        splits_config = transform_pb2.SplitsConfig()
        assert TRAIN_SPLITS in args[SPLIT_MAPPING], \
            f'When you are defining a custom split mapping, please define ' \
            f'{TRAIN_SPLITS}!'
        for process, splits in args[SPLIT_MAPPING].items():
            for split in splits:
                if process == TRAIN_SPLITS:
                    splits_config.analyze.append(split)
                splits_config.transform.append(split)
        return splits_config
    else:
        return None


class BasePreprocesserStep(BaseStep):
    """
    Base class for all preprocessing steps. These steps are used to
    specify transformation and filling operations on data that occur before
    the machine learning model is trained.
    """

    STEP_TYPE = StepTypes.preprocesser.name

    def __init__(self,
                 split_mapping: Dict[Text, List[Text]] = None,
                 **kwargs):
        """
        Base preprocessing step constructor. Custom preprocessing steps need
        to override the `preprocessing_fn` class method.

        Args:
            **kwargs: Additional keyword arguments.
        """

        super(BasePreprocesserStep, self).__init__(split_mapping=split_mapping,
                                                   **kwargs)

    def preprocessing_fn(self, inputs: Dict):
        """
        Function used in the Transform component. Override this to do custom
        preprocessing logic.

        Args:
            inputs (dict): Inputs where keys are feature names and values are
            tensors which represent the values of the features.

        Returns:
            outputs (dict): Inputs where keys are transformed feature names
             and values are tensors with the transformed values of the
             features.
        """
        pass



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""
These utils are predicated on the following definitions:

* class_source: This is a python-import type path to a class, e.g.
some.mod.class
* module_source: This is a python-import type path to a module, e.g. some.mod
* file_path, relative_path, absolute_path: These are file system paths.
* source: This is a class_source or module_source. If it is a class_source, it
can also be optionally pinned.
* pin: Whatever comes after the `@` symbol from a source, usually the git sha
or the version of zenml as a string.
"""
import importlib
import inspect
from typing import Text, Type, Optional, Union

from tfx.utils.import_utils import import_class_by_path

from zenml.constants import APP_NAME
from zenml.logger import get_logger
from zenml.repo import GitWrapper
from zenml.repo.repo import Repository
from zenml.version import __version__

logger = get_logger(__name__)


def is_standard_pin(pin: Text) -> bool:
    """
    Returns True if pin is valid ZenML pin, else False.

    Args:
        pin (str): potential ZenML pin like 'zenml_0.1.1'
    """
    if pin.startswith(f'{APP_NAME}_'):
        return True
    return False


def create_zenml_pin():
    """Creates a ZenML pin for source pinning from release version."""
    return f'{APP_NAME}_{__version__}'


def resolve_standard_source(source: Text) -> Text:
    """
    Creates a ZenML pin for source pinning from release version.

    Args:
        source (str): class_source e.g. this.module.Class.
    """
    if '@' in source:
        raise AssertionError(f'source {source} is already pinned.')
    pin = create_zenml_pin()
    return f'{source}@{pin}'


def is_standard_source(source: Text) -> bool:
    """
    Returns True if source is a standard ZenML source.

    Args:
        source (str): class_source e.g. this.module.Class[@pin].
    """
    if source.split('.')[0] == 'zenml':
        return True
    return False


def is_valid_source(source: Text) -> bool:
    """
    Checks whether the source_path is valid or not.

    Args:
        source (str): class_source e.g. this.module.Class[@pin].
    """
    try:
        load_source_path_class(source)
    except:
        return False
    return True


def get_path_from_source(source):
    """
    Get file path from source

    Args:
        source (str): class_source e.g. this.module.Class.
    """
    # TODO: [MEDIUM] Make sure this is an absolute path rather than naive split
    file_path = "/".join(source.split('.')[:-1]) + '.py'
    return file_path


def get_pin_from_source(source: Text) -> Optional[Text]:
    """
    Gets pin from source, i.e. module.path@pin, returns pin.

    Args:
        source (str): class_source e.g. this.module.Class[@pin].
    """
    if '@' in source:
        pin = source.split('@')[-1]
        return pin
    return 'unpinned'


def get_class_source_from_source(source: Text) -> Optional[Text]:
    """
    Gets class source from source, i.e. module.path@version, returns version.

    Args:
        source: source pointing to potentially pinned sha.
    """
    # source need not even be pinned
    return source.split('@')[0]


def get_module_source_from_source(source: Text) -> Text:
    """
    Gets module source from source. E.g. `some.module.file.class@version`,
    returns `some.module`.

    Args:
        source: source pointing to potentially pinned sha.
    """
    class_source = get_class_source_from_source(source)
    return '.'.join(class_source.split('.')[:-2])


def get_module_source_from_file_path(file_path):
    """
    Gets module_source from a file_path. E.g. `/home/myrepo/step/trainer.py`
    returns `myrepo.step.trainer` if `myrepo` is the root of the repo.

    Args:
        file_path: Absolute file path to a file within the module.
    """
    repo_path = Repository.get_instance().path

    # Replace repo_path with file_path to get relative path left over
    relative_file_path = file_path.replace(repo_path, '')[1:]

    # Kick out the .py and replace `/` with `.` to get the module source
    relative_file_path = relative_file_path.replace('.py', '')
    module_source = relative_file_path.replace('/', '.')
    return module_source


def get_relative_path_from_module_source(module_source: Text):
    """
    Get a directory path from module, relative to root of repository.

    E.g. zenml.core.step will return zenml/core/step.

    Args:
        module_source (str): A module e.g. zenml.core.step
    """
    return module_source.replace('.', '/')


def get_absolute_path_from_module_source(module: Text):
    """
    Get a directory path from module source.

    E.g. `zenml.core.step` will return `full/path/to/zenml/core/step`.

    Args:
        module (str): A module e.g. `zenml.core.step`.
    """
    mod = importlib.import_module(module)
    return mod.__path__[0]


def get_module_source_from_class(class_: Union[Type, Text]) -> Optional[Text]:
    """
    Takes class input and returns module_source. If class is already string
    then returns the same.

    Args:
        class_: object of type class.
    """
    if type(class_) == str:
        module_source = class_
    else:
        # Infer it from the class provided
        if not inspect.isclass(class_):
            raise Exception('step_type is neither string nor class.')
        module_source = class_.__module__ + '.' + class_.__name__
    return module_source


def resolve_class(class_: Type) -> Text:
    """
    Resolves
    Args:
        class_: A Python Class reference.

    Returns: source_path e.g. this.module.Class[@pin].
    """
    initial_source = class_.__module__ + '.' + class_.__name__
    if is_standard_source(initial_source):
        return resolve_standard_source(initial_source)

    # Get the full module path relative to the repository
    file_path = inspect.getfile(class_)
    module_source = get_module_source_from_file_path(file_path)

    class_source = module_source + '.' + class_.__name__
    return resolve_class_source(class_source)


def resolve_class_source(class_source: Text) -> Text:
    """
    Resolves class_source with an optional pin.

    Args:
        class_source (str): class_source e.g. this.module.Class
    """
    if '@' in class_source:
        # already pinned
        return class_source

    if is_standard_source(class_source):
        # that means use standard version
        return resolve_standard_source(class_source)

    # otherwise use Git resolution
    wrapper: GitWrapper = Repository.get_instance().get_git_wrapper()
    resolved_source = wrapper.resolve_class_source(class_source)
    return resolved_source


def load_source_path_class(source: Text) -> Type:
    """
    Loads a Python class from the source.

    Args:
        source (str): class_source e.g. this.module.Class[@sha]
    """
    source = source.split('@')[0]
    pin = source.split('@')[-1]
    is_standard = is_standard_pin(pin)

    if '@' in source and not is_standard:
        logger.debug('Pinned step found with git sha. '
                     'Loading class from git history.')
        wrapper: GitWrapper = Repository.get_instance().get_git_wrapper()

        module_source = get_module_source_from_source(source)
        relative_module_path = get_relative_path_from_module_source(
            module_source)

        logger.warning('Found source with a pinned sha. Will now checkout '
                       f'module: {module_source}')

        # critical step
        if not wrapper.check_module_clean(source):
            raise Exception(f'One of the files at {relative_module_path} '
                            f'is not committed and we '
                            f'are trying to load that directory from git '
                            f'history due to a pinned step in the pipeline. '
                            f'Please commit the file and then run the '
                            f'pipeline.')

        # Check out the directory at that sha
        wrapper.checkout(sha_or_branch=pin, directory=relative_module_path)

        # After this point, all exceptions will first undo the above
        try:
            class_ = import_class_by_path(source)
            wrapper.reset(relative_module_path)
            wrapper.checkout(directory=relative_module_path)
        except Exception:
            wrapper.reset(relative_module_path)
            wrapper.checkout(directory=relative_module_path)
            raise Exception
    elif '@' in source and is_standard:
        logger.debug(f'Default {APP_NAME} class used. Loading directly.')
        # TODO: [LOW] Check if ZenML version is installed before loading.
        class_ = import_class_by_path(source)
    else:
        logger.debug('Unpinned step found with no git sha. Attempting to '
                     'load class from current repository state.')
        class_ = import_class_by_path(source)

    return class_



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
import random
from pathlib import Path

import pandas as pd
import pytest

import zenml
from zenml import exceptions
from zenml.datasources import BaseDatasource
from zenml.pipelines import BasePipeline
from zenml.standards import standard_keys as keys
from zenml.utils import yaml_utils

# Nicholas a way to get to the root
ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")


def test_datasource_create(repo):
    name = "my_datasource"
    first_ds = BaseDatasource(name=name)

    assert not first_ds._immutable

    # reload a datasource from a saved config
    p_config = random.choice(repo.get_pipeline_file_paths())
    cfg = yaml_utils.read_yaml(p_config)
    second_ds = BaseDatasource.from_config(cfg[keys.GlobalKeys.PIPELINE])

    assert second_ds._immutable


def test_to_from_config(equal_datasources):
    first_ds = BaseDatasource(name="my_datasource")

    config = dict({keys.PipelineKeys.STEPS: {}})
    config[keys.PipelineKeys.STEPS][keys.DataSteps.DATA] = {"args": {}}
    config[keys.PipelineKeys.DATASOURCE] = first_ds.to_config()
    second_ds = BaseDatasource.from_config(config)

    assert equal_datasources(first_ds, second_ds, loaded=True)


def test_get_one_pipeline(repo):
    name = "my_datasource"
    first_ds = BaseDatasource(name=name)

    with pytest.raises(exceptions.EmptyDatasourceException):
        _ = first_ds._get_one_pipeline()

    # reload a datasource from a saved config
    p_config = random.choice(repo.get_pipeline_file_paths())
    cfg = yaml_utils.read_yaml(p_config)
    second_ds = BaseDatasource.from_config(cfg[keys.GlobalKeys.PIPELINE])

    assert second_ds._get_one_pipeline()


def test_get_data_file_paths(repo):
    first_ds = BaseDatasource(name="my_datasource")

    first_pipeline = BasePipeline(name="my_pipeline")

    first_pipeline.add_datasource(first_ds)

    # reload a datasource from a saved config
    p_config = random.choice(repo.get_pipeline_file_paths())
    cfg = yaml_utils.read_yaml(p_config)
    second_ds = BaseDatasource.from_config(cfg[keys.GlobalKeys.PIPELINE])

    with pytest.raises(AssertionError):
        _ = second_ds._get_data_file_paths(first_pipeline)

    real_pipeline = second_ds._get_one_pipeline()
    paths = second_ds._get_data_file_paths(real_pipeline)

    # TODO: Find a better way of asserting TFRecords
    assert all(os.path.splitext(p)[-1] == ".gz" for p in paths)


def test_get_datapoints(repo):
    # reload a datasource from a saved config
    p_config = random.choice(repo.get_pipeline_file_paths())
    cfg = yaml_utils.read_yaml(p_config)
    ds: BaseDatasource = BaseDatasource.from_config(
        cfg[keys.GlobalKeys.PIPELINE])

    csv_df = pd.read_csv(os.path.join(TEST_ROOT,
                                      "test_data", "my_dataframe.csv"))

    assert ds.n_datapoints == len(csv_df.index)


def test_sample_data(repo):
    # reload a datasource from a saved config
    p_config = random.choice(repo.get_pipeline_file_paths())
    cfg = yaml_utils.read_yaml(p_config)
    ds: BaseDatasource = BaseDatasource.from_config(
        cfg[keys.GlobalKeys.PIPELINE])

    sample_df = ds.sample_data()
    csv_df = pd.read_csv(os.path.join(TEST_ROOT,
                                      "test_data", "my_dataframe.csv"))

    # TODO: This fails for floating point values other than 2.5 in GPA.
    #   Pandas floating point comp might be too strict
    assert sample_df.equals(csv_df)



import os

from zenml.backends.processing import ProcessingDataFlowBackend
from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository, ArtifactStore
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.utils.naming_utils import transformed_label_name

GCP_PROJECT = os.getenv('GCP_PROJECT')
GCP_BUCKET = os.getenv('GCP_BUCKET')
assert GCP_BUCKET
assert GCP_PROJECT

# Run the pipeline locally but distribute the beam-compatible steps, i.e.,
# the Data, Statistics, Preprocessing and Evaluator Steps.
# Note: If any of these steps are non-standard, custom steps, then you need
# to build a new Docker image based on the ZenML Dataflow image, and pass that
# into the `image` parameter in the ProcessingDataFlowBackend

# Define the processing backend
processing_backend = ProcessingDataFlowBackend(
    project=GCP_PROJECT,
    staging_location=os.path.join(GCP_BUCKET, 'dataflow_processing/staging'),
)

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(
    RandomSplit(split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}).with_backend(
        processing_backend)
)

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ).with_backend(processing_backend)
)

# Add a trainer
training_pipeline.add_trainer(TFFeedForwardTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    output_units=1,
    metrics=['accuracy'],
    epochs=20))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(
        slices=[['has_diabetes']],
        metrics={transformed_label_name('has_diabetes'):
                     ['binary_crossentropy', 'binary_accuracy']}
    ).with_backend(processing_backend)
)
# Define the artifact store
artifact_store = ArtifactStore(
    os.path.join(GCP_BUCKET, 'dataflow_processing/artifact_store'))

# Run the pipeline
training_pipeline.run(artifact_store=artifact_store)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from pathlib import Path

import zenml
from zenml.datasources.json_datasource import JSONDatasource

ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")


def test_json_commit(repo):
    data = Path(__file__).resolve() \
               .parent.parent.parent / 'tests' / 'test_data' / 'simple.json'
    ds = JSONDatasource(
        name='json_ds',
        path=str(data.resolve())
    )

    assert not ds.commits

    _id = ds.commit()

    # check that this commit made
    assert _id == ds.get_latest_commit()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint for gcp orchestrator"""

import base64
import json
import os
import sys

import fire

from zenml.backends.orchestrator.base.orchestrator_base_backend import \
    OrchestratorBaseBackend
from zenml.backends.orchestrator.gcp.orchestrator_gcp_backend import \
    TAR_PATH_ARG
from zenml.repo.repo import Repository
from zenml.standards import standard_keys as keys
from zenml.utils import path_utils

EXTRACTED_TAR_DIR_NAME = 'zenml_working.tar.gz'
EXTRACTED_TAR_FILE_PATH = os.path.join(os.getcwd(), EXTRACTED_TAR_DIR_NAME)
EXTRACTED_TAR_DIR = os.path.join(os.getcwd(), 'zenml_working')


class PipelineRunner(object):

    def run_pipeline(self, config_b64: str):
        # Load config from base64
        config = json.loads(base64.b64decode(config_b64))

        # Remove tar_path arg from config
        tar_path = config[keys.GlobalKeys.BACKEND][keys.BackendKeys.ARGS].pop(
            TAR_PATH_ARG)

        # Copy it over locally because it will be remote
        path_utils.copy(tar_path, EXTRACTED_TAR_FILE_PATH)

        # Extract it to EXTRACTED_TAR_DIR
        path_utils.extract_tarfile(EXTRACTED_TAR_FILE_PATH, EXTRACTED_TAR_DIR)

        # Append to sys to make user code discoverable
        sys.path.append(EXTRACTED_TAR_DIR)

        # Make sure the Repository is initialized at the right path
        Repository.get_instance(EXTRACTED_TAR_DIR)

        # Change orchestrator of pipeline to local
        OrchestratorBaseBackend().run(config)


if __name__ == "__main__":
    # execute only if run as a script
    fire.Fire(PipelineRunner)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from enum import Enum


class PipelineStatusTypes(Enum):
    NotStarted = 1
    Failed = 2
    Succeeded = 3
    Running = 4


class GDPComponent(Enum):
    SplitGen = 1
    SplitStatistics = 2
    SplitSchema = 3
    Sequencer = 4
    SequencerStatistics = 5
    SequencerSchema = 6
    PreTransform = 7
    PreTransformStatistics = 8
    PreTransformSchema = 9
    Transform = 10
    Trainer = 11
    Evaluator = 12
    ResultPackager = 13
    ModelValidator = 14
    Deployer = 15
    DataGen = 16
    Inferrer = 17
    DataStatistics = 18
    DataSchema = 19
    Tokenizer = 20


class MLMetadataTypes(Enum):
    sqlite = 1
    mysql = 2
    mock = 3


class ArtifactStoreTypes(Enum):
    local = 1
    gcs = 2


class StepTypes(Enum):
    base = 1
    data = 2
    sequencer = 3
    preprocesser = 4
    split = 5
    trainer = 6
    evaluator = 7
    deployer = 8
    inferrer = 9


class GCPGPUTypes(Enum):
    K80 = 1
    V100 = 2
    P100 = 3


class ImagePullPolicy(Enum):
    Always = 1
    Never = 2
    IfNotPresent = 3



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Split step utils tests."""

import pytest
import tensorflow as tf

from zenml.steps.split.utils import get_categorical_value, \
    partition_cat_list


def test_get_categorical_value():

    feature = {"int64cat_col":
               tf.train.Feature(int64_list=tf.train.Int64List(value=[1])),
               "floatcat_col":
               tf.train.Feature(float_list=tf.train.FloatList(value=[42.0])),
               "bytescat_col":
               tf.train.Feature(bytes_list=tf.train.BytesList(
                   value=[b"test_string"]))
               }

    tf_example = tf.train.Example(features=tf.train.Features(feature=feature))

    int_value = get_categorical_value(tf_example, cat_col="int64cat_col")
    float_value = get_categorical_value(tf_example, cat_col="floatcat_col")
    str_value = get_categorical_value(tf_example, cat_col="bytescat_col")

    # ensure type consistency
    assert isinstance(int_value, int)
    assert isinstance(float_value, float)
    assert isinstance(str_value, str)

    # tests for a categorical feature that is not in the data
    with pytest.raises(AssertionError):
        _ = get_categorical_value(tf_example, cat_col="testest")


def test_partition_cat_list():
    cat_list = [f"cat{i}" for i in range(10)]

    c_ratio = {"train": 0.5,
               "eval": 0.3,
               "test": 0.2}

    cat_dict = partition_cat_list(cat_list=cat_list, c_ratio=c_ratio)

    assert [len(li) for li in cat_dict.values()] == [5, 3, 2]

    # test: two categories, but three folds
    cat_list = [f"cat{i}" for i in range(2)]

    c_ratio = {"train": 0.33,
               "eval": 0.33,
               "test": 0.34}

    cat_dict = partition_cat_list(cat_list=cat_list, c_ratio=c_ratio)

    # expected: train and eval each get one, test empty
    assert [len(li) for li in cat_dict.values()] == [1, 0, 1]

    # test: three categories and three folds
    cat_list = [f"cat{i}" for i in range(3)]

    c_ratio = {"train": 0.33,
               "eval": 0.33,
               "test": 0.34}

    cat_dict = partition_cat_list(cat_list=cat_list, c_ratio=c_ratio)

    # expected: train, eval, test each get one
    assert [len(li) for li in cat_dict.values()] == [1, 1, 1]



from __future__ import division

from abc import ABC
from itertools import starmap

import numpy as np
import tensorflow as tf
import torch
import torch.utils.data as data
from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.data.ops import readers as core_readers
from tensorflow.python.framework import dtypes

from zenml.utils import naming_utils


class TFRecordTorchDataset(data.IterableDataset, ABC):
    def __init__(self,
                 file_pattern,
                 spec) -> None:
        super(TFRecordTorchDataset, self).__init__()

        self.dataset = create_tf_dataset(file_pattern=file_pattern,
                                         spec=spec)

    def __iter__(self):
        it = _create_iterator(self.dataset)
        it = _shuffle_iterator(it, 12)
        return it


def _create_iterator(dataset):
    iterator = dataset.as_numpy_iterator()
    iterator = starmap(_convert_to_tensors, iterator)
    return iterator


def _convert_to_tensors(features, label, raw):
    return {k: torch.from_numpy(v) for k, v in features.items()}, \
           {k: torch.from_numpy(v) for k, v in label.items()}, \
           raw


def _shuffle_iterator(iterator,
                      queue_size: int):
    buffer = []
    try:
        for _ in range(queue_size):
            buffer.append(next(iterator))
    except StopIteration:
        raise Exception('Iteration stopped!')
    while buffer:
        index = np.random.randint(len(buffer))
        try:
            item = buffer[index]
            buffer[index] = next(iterator)
            yield item
        except StopIteration:
            yield buffer.pop(index)


def _gzip_reader_fn(filenames):
    """
    Small utility returning a record reader that can read gzipped files.

    Args:
        filenames: Names of the compressed TFRecord data files.
    """
    return tf.data.TFRecordDataset(filenames, compression_type='GZIP')


def _split_inputs_labels(x):
    inputs = {}
    labels = {}
    raw = {}
    for e in x:
        if naming_utils.check_if_transformed_label(e):
            labels[e] = x[e]
        elif naming_utils.check_if_transformed_feature(e):
            inputs[e] = x[e]
        else:
            raw[e] = x[e]

    return inputs, labels, raw


def create_tf_dataset(file_pattern,
                      spec,
                      num_epochs=1,
                      shuffle=False,
                      shuffle_seed=None,
                      shuffle_buffer_size=None,
                      reader_num_threads=None,
                      prefetch_buffer_size=None):
    reader = _gzip_reader_fn

    if reader_num_threads is None:
        reader_num_threads = 1
    if prefetch_buffer_size is None:
        prefetch_buffer_size = dataset_ops.AUTOTUNE

    dataset = dataset_ops.Dataset.list_files(file_pattern=file_pattern,
                                             shuffle=shuffle,
                                             seed=shuffle_seed)

    if reader_num_threads == dataset_ops.AUTOTUNE:
        dataset = dataset.interleave(lambda filename: reader(filename),
                                     num_parallel_calls=reader_num_threads)
        options = dataset_ops.Options()
        options.experimental_deterministic = True
        dataset = dataset.with_options(options)
    else:
        def apply_fn(d):
            return core_readers.ParallelInterleaveDataset(
                d,
                lambda filename: reader(filename),
                cycle_length=reader_num_threads,
                block_length=1,
                sloppy=True,
                buffer_output_elements=None,
                prefetch_input_elements=None)

        dataset = dataset.apply(apply_fn)

    if dataset_ops.get_legacy_output_types(dataset) == (dtypes.string,
                                                        dtypes.string):
        dataset = dataset_ops.MapDataset(dataset,
                                         lambda _, v: v,
                                         use_inter_op_parallelism=False)

    if shuffle:
        dataset = dataset.shuffle(shuffle_buffer_size, shuffle_seed)
    if num_epochs != 1:
        dataset = dataset.repeat(num_epochs)

    dataset = dataset.map(lambda x: tf.io.parse_example(x, spec))
    dataset = dataset.map(_split_inputs_labels)
    dataset = dataset.prefetch(prefetch_buffer_size)
    return dataset



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML NLP Pipeline Prototype."""

import os
from typing import Dict, Text, Any, List
from typing import Optional, Union

import tensorflow as tf
from tfx.components.common_nodes.importer_node import ImporterNode
from tfx.components.schema_gen.component import SchemaGen
from tfx.components.statistics_gen.component import StatisticsGen
from tfx.components.trainer.component import Trainer
from tfx.proto import trainer_pb2
from tfx.types import standard_artifacts

from zenml import constants
from zenml.backends.training import TrainingBaseBackend
from zenml.components import SplitGen, Trainer
from zenml.components import Tokenizer
from zenml.enums import GDPComponent
from zenml.enums import PipelineStatusTypes
from zenml.logger import get_logger
from zenml.pipelines import BasePipeline
from zenml.standards import standard_keys as keys
from zenml.steps.split import BaseSplitStep
from zenml.steps.tokenizer import BaseTokenizer
from zenml.steps.trainer import BaseTrainerStep

logger = get_logger(__name__)


class NLPPipeline(BasePipeline):
    PIPELINE_TYPE = "nlp"

    def predict_sentence(self, sequence: Union[Text, List[Text]]):
        """Call operator for local inference method"""

        if not self.get_status() == PipelineStatusTypes.Succeeded.name:
            logger.info(
                "Please run the pipeline first before running inference!")
            return

        trainer_step = self.steps_dict[keys.NLPSteps.TRAINER]

        # e.g. HuggingFace has special APIs for model loading,
        # so we fall back to a trainer step class method
        model_uri = os.path.join(self.get_model_uri(), "serving_model_dir")
        model = trainer_step.load_model(model_uri)

        tokenizer_step = self.steps_dict[keys.NLPSteps.TOKENIZER]

        tokenizer_step.load_vocab(self.get_tokenizer_uri())

        encoded = tokenizer_step.encode(sequence=sequence,
                                        output_format="tf_tensors")

        transformed_bert_features = {k: tf.reshape(v, (1, -1)) for k, v in
                                     encoded.items()}

        prediction = model(input_ids=transformed_bert_features, training=False)

        formatted = [
            {"label": model.config.id2label[item.argmax()],
             "score": item.max().item()}
            for item in tf.math.sigmoid(prediction.logits).numpy()
        ]

        logger.info(formatted)

    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Builds the NLP pipeline as a series of TFX components.

        Args:
            config: A ZenML configuration in dictionary format.

        Returns:
            A chronological list of TFX components making up the NLP
             pipeline.

        """
        steps = config[keys.GlobalKeys.PIPELINE][keys.PipelineKeys.STEPS]

        component_list = []

        ############
        # RAW DATA #
        ############
        # if there are no commits, then lets make one
        if self.datasource.is_empty:
            logger.info(
                f'Datasource {self.datasource.name} has no commits. Creating '
                f'the first one..')
            self.datasource_commit_id = self.datasource.commit()

        data_pipeline = self.datasource.get_data_pipeline_from_commit(
            self.datasource_commit_id)

        data = ImporterNode(
            instance_name=GDPComponent.DataGen.name,
            source_uri=data_pipeline.get_artifacts_uri_by_component(
                GDPComponent.DataGen.name)[0],
            artifact_type=standard_artifacts.Examples)
        component_list.extend([data])

        #############
        # TOKENIZER #
        #############
        tokenizer_config = steps[keys.NLPSteps.TOKENIZER]
        tokenizer = Tokenizer(
            source=tokenizer_config[keys.StepKeys.SOURCE],
            source_args=tokenizer_config[keys.StepKeys.ARGS],
            examples=data.outputs.result,
        ).with_id(GDPComponent.Tokenizer.name)

        component_list.extend([tokenizer])

        # return component_list

        statistics_data = StatisticsGen(
            examples=tokenizer.outputs.output_examples
        ).with_id(GDPComponent.DataStatistics.name)

        schema_data = SchemaGen(
            statistics=statistics_data.outputs.output,
            infer_feature_shape=True,
        ).with_id(GDPComponent.DataSchema.name)

        split_config = steps[keys.NLPSteps.SPLIT]
        splits = SplitGen(
            input_examples=tokenizer.outputs.output_examples,
            source=split_config[keys.StepKeys.SOURCE],
            source_args=split_config[keys.StepKeys.ARGS],
            schema=schema_data.outputs.schema,
            statistics=statistics_data.outputs.output,
        ).with_id(GDPComponent.SplitGen.name)

        component_list.extend([data,
                               statistics_data,
                               schema_data,
                               splits])

        ############
        # TRAINING #
        ############
        training_backend: Optional[TrainingBaseBackend] = \
            self.steps_dict[keys.NLPSteps.TRAINER].backend

        # default to local
        if training_backend is None:
            training_backend = TrainingBaseBackend()

        training_kwargs = {
            'custom_executor_spec': training_backend.get_executor_spec(),
            'custom_config': steps[keys.NLPSteps.TRAINER]
        }
        training_kwargs['custom_config'].update(
            training_backend.get_custom_config())

        trainer = Trainer(
            examples=splits.outputs.examples,
            run_fn=constants.TRAINER_FN,
            schema=schema_data.outputs.schema,
            **training_kwargs
        ).with_id(GDPComponent.Trainer.name)

        component_list.extend([trainer])

        return component_list

    def steps_completed(self) -> bool:
        mandatory_steps = [keys.NLPSteps.TOKENIZER, keys.NLPSteps.SPLIT,
                           keys.NLPSteps.TRAINER]

        for step_name in mandatory_steps:
            if step_name not in self.steps_dict.keys():
                raise AssertionError(f'Mandatory step {step_name} not added.')
        return True

    def add_tokenizer(self, tokenizer_step: BaseTokenizer):
        self.steps_dict[keys.NLPSteps.TOKENIZER] = tokenizer_step

    def add_split(self, split_step: BaseSplitStep):
        self.steps_dict[keys.NLPSteps.SPLIT] = split_step

    def add_trainer(self, trainer_step: BaseTrainerStep):
        self.steps_dict[keys.NLPSteps.TRAINER] = trainer_step

    def get_model_uri(self):
        """Gets model artifact."""
        uris = self.get_artifacts_uri_by_component(
            GDPComponent.Trainer.name, False)
        return uris[0]

    def get_schema_uri(self):
        """Gets transform artifact."""
        uris = self.get_artifacts_uri_by_component(
            GDPComponent.DataSchema.name, False)
        return uris[0]

    def get_tokenizer_uri(self):
        """Gets transform artifact."""
        uris = self.get_artifacts_uri_by_component(
            GDPComponent.Tokenizer.name, False)
        return uris[0]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Wrapper class to handle Git integration"""

import os
from typing import Text, List

from git import Repo as GitRepo, BadName

from zenml.logger import get_logger
from zenml.repo.constants import GIT_FOLDER_NAME
from zenml.utils import path_utils

logger = get_logger(__name__)


class GitWrapper:
    """Wrapper class for Git.

    This class is responsible for handling git interactions, primarily
    handling versioning of different steps in pipelines.
    """

    def __init__(self, repo_path: Text):
        """
        Initialize GitWrapper. Should be initialize by ZenML Repository.
        Args:
            repo_path:

        Raises:
            InvalidGitRepositoryError: If repository is not a git repository.
            NoSuchPathError: If the repo_path does not exist.
        """
        # TODO: [LOW] Raise ZenML exceptions here instead.
        self.repo_path: Text = repo_path
        self.git_root_path: Text = os.path.join(repo_path, GIT_FOLDER_NAME)
        self.git_repo: GitRepo = GitRepo(self.repo_path)

        # TODO: [LOW] Check whether this is necessary.
        assert not self.git_repo.bare

    def add_gitignore(self, items: List[Text]):
        """
        Adds `items` to .gitignore, if .gitignore exists. Otherwise creates
        and adds.

        Args:
            items (list[str]): Items to add.
        """
        str_items = '\n'.join(items)
        str_items = '\n\n# ZenML\n' + str_items

        gitignore_path = os.path.join(self.repo_path, '.gitignore')
        if not path_utils.file_exists(gitignore_path):
            path_utils.create_file_if_not_exists(gitignore_path, str_items)
        else:
            path_utils.append_file(gitignore_path, str_items)

    def check_file_committed(self, file_path: Text) -> bool:
        """
        Checks file is committed. If yes, return True, else False.

        Args:
            file_path (str): Path to any file within the ZenML repo.
        """
        uncommitted_files = [i.a_path for i in self.git_repo.index.diff(None)]
        try:
            staged_files = [i.a_path for i in self.git_repo.index.diff('HEAD')]
        except BadName:
            # for Ref 'HEAD' did not resolve to an object
            logger.debug('No committed files in the repo. No staged files.')
            staged_files = []

        # source: https://stackoverflow.com/questions/3801321/
        untracked_files = self.git_repo.git.ls_files(
            others=True, exclude_standard=True).split('\n')
        for item in uncommitted_files + staged_files + untracked_files:
            # These are all changed files
            if file_path == item:
                return False
        return True

    def get_current_sha(self) -> Text:
        """
        Finds the git sha that each file within the module is currently on.
        """
        return self.git_repo.head.object.hexsha

    def check_module_clean(self, source: Text):
        """
        Returns True if all files within source's module are committed.

        Args:
            source (str): relative module path pointing to a Class.
        """
        # import here to resolve circular dependency
        from zenml.utils import source_utils

        # Get the module path
        module_path = source_utils.get_module_source_from_source(source)

        # Get relative path of module because check_file_committed needs that
        module_dir = source_utils.get_relative_path_from_module_source(
            module_path)

        # Get absolute path of module because path_utils.list_dir needs that
        mod_abs_dir = source_utils.get_absolute_path_from_module_source(
            module_path)
        module_file_names = path_utils.list_dir(
            mod_abs_dir, only_file_names=True)

        # Go through each file in module and see if there are uncommitted ones
        for file_path in module_file_names:
            path = os.path.join(module_dir, file_path)

            # if its .gitignored then continue and dont do anything
            if len(self.git_repo.ignored(path)) > 0:
                continue

            if path_utils.is_dir(os.path.join(mod_abs_dir, file_path)):
                logger.warning(
                    f'The step {source} is contained inside a module '
                    f'that '
                    f'has sub-directories (the sub-directory {file_path} at '
                    f'{mod_abs_dir}). For now, ZenML supports only a flat '
                    f'directory structure in which to place Steps. Please make'
                    f' sure that the Step does not utilize the sub-directory.')
            if not self.check_file_committed(path):
                return False
        return True

    def resolve_class_source(self, source: Text) -> Text:
        """
        Takes source (e.g. this.module.ClassName), and appends relevant
        sha to it if the files within `module` are all committed. If even one
        file is not committed, then returns `source` unchanged.

        Args:
            source (str): relative module path pointing to a Class.
        """
        if not self.check_module_clean(source):
            # Return the source path if not clean
            logger.warning(
                f'Found uncommitted file. Pipelines run with this '
                f'configuration may not be reproducible. Please commit '
                f'all files in this module and then run the pipeline to '
                f'ensure reproducibility.')
            return source
        return source + '@' + self.get_current_sha()

    def stash(self):
        """Wrapper for git stash"""
        git = self.git_repo.git
        git.stash()

    def stash_pop(self):
        """Wrapper for git stash pop. Only pops if there's something to pop."""
        git = self.git_repo.git
        if git.stash('list') != '':
            git.stash('pop')

    def checkout(self, sha_or_branch: Text = None, directory: Text = None):
        """
        Wrapper for git checkout

        Args:
            sha_or_branch: hex string of len 40 representing git sha OR
            name of branch
            directory (str): relative path to directory to scope checkout
        """
        # TODO: [MEDIUM] Implement exception handling
        git = self.git_repo.git
        if sha_or_branch is None:
            # Checks out directory at sha_or_branch
            assert directory is not None
            git.checkout('--', directory)
        elif directory is not None:
            assert sha_or_branch is not None
            # Basically discards all changes in directory
            git.checkout(sha_or_branch, '--', directory)
        else:
            # The case where sha_or_branch is not None and directory is None
            # In this case, the whole repo is checked out at sha_or_branch
            git.checkout(sha_or_branch)

    def reset(self, directory: Text = None):
        """
        Wrapper for `git reset HEAD <directory>`.

        Args:
            directory (str): relative path to directory to scope checkout
        """
        git = self.git_repo.git
        git.reset('HEAD', directory)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Analytics code for ZenML"""

import platform
import sys
from typing import Text

import analytics
import distro
import requests

from zenml.constants import IS_DEBUG_ENV
from zenml.logger import get_logger
from zenml.version import __version__

logger = get_logger(__name__)

# EVENTS

# Datasources

GET_DATASOURCES = "Datasources listed"

CREATE_DATASOURCE = "Datasource created"

# Functions

CREATE_STEP = "Step created"

GET_STEPS_VERSIONS = "Step Versions listed"

GET_STEP_VERSION = "Step listed"

# Pipelines

CREATE_PIPELINE = "Pipeline created"

REGISTER_PIPELINE = "Pipeline registered"

RUN_PIPELINE = "Pipeline run"

GET_PIPELINES = "Pipelines fetched"

GET_PIPELINE_ARTIFACTS = "Pipeline Artifacts fetched"

# Repo

CREATE_REPO = "Repository created"

INITIALIZE = "ZenML initialized"


def get_segment_key() -> Text:
    if IS_DEBUG_ENV:
        url = 'https://zenml.io/dev.analytics.json'
    else:
        url = 'https://zenml.io/analytics.json'

    headers = {"content-type": "application/json"}

    try:
        r = requests.get(url, headers=headers, timeout=5)
        return r.json()['id']
    except requests.exceptions.RequestException:
        logger.debug("Failed to get segment write key", exc_info=True)


analytics.write_key = get_segment_key()


def get_system_info():
    system = platform.system()

    if system == "Windows":
        version = sys.getwindowsversion()

        return {
            "os": "windows",
            "windows_version_build": version.build,
            "windows_version_major": version.major,
            "windows_version_minor": version.minor,
            "windows_version_service_pack": version.service_pack,
        }

    if system == "Darwin":
        return {"os": "mac", "mac_version": platform.mac_ver()[0]}

    if system == "Linux":
        return {
            "os": "linux",
            "linux_distro": distro.id(),
            "linux_distro_like": distro.like(),
            "linux_distro_version": distro.version(),
        }

    # We don't collect data for any other system.
    return {"os": "unknown"}


def track_event(event, metadata=None):
    """
    Track segment event if user opted-in.

    Args:
        event: name of event to track in segment.
        metadata: dict of metadata
    """
    try:
        from zenml.repo import GlobalConfig
        config = GlobalConfig.get_instance()
        opt_in = config.get_analytics_opt_in()
        logger.debug(f"Analytics opt-in: {opt_in}.")

        if opt_in is False and event is not INITIALIZE:
            return

        user_id = config.get_user_id()

        if metadata is None:
            metadata = {}

        # add basics
        metadata.update(get_system_info())
        metadata.update({'version': __version__})

        analytics.track(user_id, event, metadata)
        logger.debug(
            f'Analytics sent: User: {user_id}, Event: {event}, Metadata: '
            f'{metadata}')
    except Exception as e:
        # We should never fail main thread
        logger.debug(f'Analytics failed due to: {e}')
        return


def parametrized(dec):
    """This is a meta-decorator, that is, a decorator for decorators.
    As a decorator is a function, it actually works as a regular decorator
    with arguments:"""

    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)

        return repl

    return layer


@parametrized
def track(func, event=None):
    if event is None:
        event = func.__name__  # default to name of function

    metadata = {}

    # TODO: [LOW] See if we can get anonymized data from func
    # if func.__name__:
    #     metadata['function'] = func.__name__
    # if func.__module__:
    #     metadata['module'] = func.__module__

    def inner_func(*args, **kwargs):
        track_event(event, metadata=metadata)
        result = func(*args, **kwargs)
        return result

    return inner_func



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from typing import Dict, List, Text

from zenml.enums import StepTypes
from zenml.steps import BaseStep
from zenml.steps.trainer.utils import TRAIN_SPLITS, TEST_SPLITS, EVAL_SPLITS


class BaseEvaluatorStep(BaseStep):
    """
    Base evaluator step. All evaluator steps should inherit from this class.
    """
    CUSTOM_MODULE = None

    STEP_TYPE = StepTypes.evaluator.name

    def __init__(self,
                 split_mapping: Dict[Text, List[Text]] = None,
                 **kwargs):

        if split_mapping:
            assert len(split_mapping[TRAIN_SPLITS]) > 0, \
                'While defining your own mapping, you need to provide at least ' \
                'one training split.'
            assert len(split_mapping[EVAL_SPLITS]) > 0, \
                'While defining your own mapping, you need to provide at least ' \
                'one eval split.'

            if TEST_SPLITS not in split_mapping:
                split_mapping.update({TEST_SPLITS: []})
            assert len(split_mapping) == 3, \
                f'While providing a split_mapping please only use ' \
                f'{TRAIN_SPLITS}, {EVAL_SPLITS} and {TEST_SPLITS} as keys.'

            self.split_mapping = split_mapping
        else:
            self.split_mapping = {TRAIN_SPLITS: ['train'],
                                  EVAL_SPLITS: ['eval'],
                                  TEST_SPLITS: ['test']}

        super(BaseEvaluatorStep, self).__init__(split_mapping=split_mapping,
                                                **kwargs)

    def build_config(self):
        pass



from typing import Optional
from typing import Text, Dict, Any

from tfx import types
from tfx.dsl.components.base import base_component
from tfx.dsl.components.base import executor_spec
from tfx.types.component_spec import ChannelParameter
from tfx.types.component_spec import ComponentSpec
from tfx.types.component_spec import ExecutionParameter
from tfx.types.standard_artifacts import Examples, Model, Schema, \
    ModelEvaluation

from zenml.components.evaluator import constants
from zenml.components.evaluator import executor


class ZenMLEvaluatorSpec(ComponentSpec):
    PARAMETERS = {constants.SOURCE: ExecutionParameter(type=Text),
                  constants.ARGS: ExecutionParameter(type=Dict[Text, Any])}

    INPUTS = {constants.EXAMPLES: ChannelParameter(type=Examples),
              constants.MODEL: ChannelParameter(type=Model, optional=True),
              constants.SCHEMA: ChannelParameter(type=Schema, optional=True)}

    OUTPUTS = {constants.EVALUATION: ChannelParameter(type=ModelEvaluation)}


class Evaluator(base_component.BaseComponent):
    """
    A new adapted version version of the TFX Evaluator component.

    In contrast to the original evaluator component, it utilizes a ZenML
    EvaluatorStep and allows the model agnostic evaluation.
    """
    SPEC_CLASS = ZenMLEvaluatorSpec
    EXECUTOR_SPEC = executor_spec.ExecutorClassSpec(executor.Executor)

    def __init__(
            self,
            source: Text,
            source_args: Dict[Text, Any],
            examples: types.Channel = None,
            model: types.Channel = None,
            output: Optional[types.Channel] = None,
            instance_name: Optional[Text] = None,
            schema: Optional[types.Channel] = None):

        # Create the output artifact if not provided
        evaluation = output or types.Channel(type=ModelEvaluation)

        # Create the spec
        spec = ZenMLEvaluatorSpec(source=source,
                                  args=source_args,
                                  examples=examples,
                                  model=model,
                                  schema=schema,
                                  evaluation=evaluation)
        super(Evaluator, self).__init__(spec=spec, instance_name=instance_name)



import re
import subprocess
import sys

########
# BASE #
########
BASE_REQUIREMENTS = ["absl-py==0.10.0",
                     "pip-check-reqs>=2.0.1,<3",
                     "click>=7.0,<8",
                     "setuptools>=38.4.0",
                     "nbformat>=5.0.4",
                     "panel==0.8.3",
                     "plotly==4.0.0",
                     "tabulate==0.8.7",
                     "numpy==1.18.0",
                     "httplib2==0.17.0",
                     "tfx==0.26.1",
                     "fire==0.3.1",
                     "gitpython==3.1.11",
                     "analytics-python==1.2.9",
                     "distro==1.5.0",
                     "tensorflow>=2.3.0,<2.4.0",
                     "tensorflow-serving-api==2.3.0"]

#####################
# EXTRAS: PROVIDERS #
#####################
GCP_INTEGRATION = 'gcp'
GCP_REQUIREMENTS = ["apache-beam[gcp]==2.27.0",
                    "apache-beam==2.27.0",
                    "google-apitools==0.5.31"]

AWS_INTEGRATION = 'aws'
AWS_REQUIREMENTS = ["boto3==1.16.62"]

AZURE_INTEGRATION = 'azure'
AZURE_REQUIREMENTS = []

###################
# EXTRAS: TOOLING #
###################
PYTORCH_INTEGRATION = 'pytorch'
PYTORCH_REQUIREMENTS = ['torch==1.7.0']

CORTEX_INTEGRATION = 'cortex'
CORTEX_REQUIREMENTS = ['cortex==0.29.0']

###############
# DATASOURCES #
###############
POSTGRES_INTEGRATION = 'postgres'
POSTGRES_REQUIREMENTS = ['beam-nuggets==0.17.0', 'pg8000==1.16.5',
                         'sqlalchemy==1.3.22']

#######################
# NLP via HuggingFace #
#######################
HUGGINGFACE_INTEGRATION = 'huggingface'
HUGGINGFACE_REQUIREMENTS = ['transformers==4.3.3', 'tokenizers==0.10.1']

###############
# EXTRAS: ALL #
###############
ALL_INTEGRATION = 'all'
ALL_REQUIREMENTS = BASE_REQUIREMENTS + \
                   GCP_REQUIREMENTS + \
                   PYTORCH_REQUIREMENTS + \
                   AZURE_REQUIREMENTS + \
                   AWS_REQUIREMENTS + \
                   POSTGRES_REQUIREMENTS + \
                   CORTEX_REQUIREMENTS + \
                   HUGGINGFACE_REQUIREMENTS

EXTRAS_REQUIRE = {GCP_INTEGRATION: GCP_REQUIREMENTS,
                  AWS_INTEGRATION: AWS_REQUIREMENTS,
                  # AZURE_INTEGRATION: AZURE_REQUIREMENTS,
                  PYTORCH_INTEGRATION: PYTORCH_REQUIREMENTS,
                  CORTEX_INTEGRATION: CORTEX_REQUIREMENTS,
                  POSTGRES_INTEGRATION: POSTGRES_REQUIREMENTS,
                  HUGGINGFACE_INTEGRATION: HUGGINGFACE_REQUIREMENTS,
                  ALL_INTEGRATION: ALL_REQUIREMENTS}


##################
# UTIL FUNCTIONS #
##################
def check_integration(integration):
    # Get the installed packages
    reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
    installed_packages = [r.decode().split('==')[0] for r in reqs.split()]

    # Get the required extra packages for the integration
    assert integration in EXTRAS_REQUIRE, \
        f'At this moment, there is no integration for {integration}. ' \
        f'Possible integrations for ZenML ' \
        f'include: {list(EXTRAS_REQUIRE.keys())}.'

    specs = EXTRAS_REQUIRE[integration]

    for s in specs:
        # Decouple from the version
        pattern = r"([a-zA-Z0-9\-]+)(\[.+\])*(.*)"
        s = re.search(pattern, s)[1]

        # TODO: We can also validate the version
        if s not in installed_packages:
            raise ModuleNotFoundError(
                f"{integration} integration not installed. "
                f"Please install zenml[{integration}] via "
                f"`pip install zenml[{integration}]`")


def list_integrations():
    """Prints integrations in an easy to read format."""
    for k, v in EXTRAS_REQUIRE.items():
        from zenml.logger import get_logger
        logger = get_logger(__name__)
        logger.info("*********")
        logger.info(f"The integration {k} has the following dependencies: {v}")
        logger.info(f'To install: pip install zenml[{k}]')



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the Kubernetes Orchestrator Backend"""

import base64
import json
import os
import time
from typing import Dict, Any, Text

from kubernetes import client as k8s_client
from kubernetes import config as k8s_config
from kubernetes.config.config_exception import ConfigException

from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.repo import Repository
from zenml.standards import standard_keys as keys
from zenml.utils import path_utils
from zenml.constants import ZENML_BASE_IMAGE_NAME, K8S_ENTRYPOINT
from zenml.enums import ImagePullPolicy
from zenml.logger import get_logger
from zenml.utils.string_utils import to_dns1123, get_id

logger = get_logger(__name__)

EXTRACTED_TAR_DIR_NAME = 'zenml_working'
TAR_PATH_ARG = 'tar_path'
STAGING_AREA = 'staging'
DEFAULT_K8S_CONFIG = os.path.join(os.path.expanduser('~'), '.kube/config')


class OrchestratorKubernetesBackend(OrchestratorBaseBackend):
    """
    Runs pipeline on a Kubernetes cluster.

    This orchestrator creates a .tar.gz of the current ZenML repository, sends
    it over to the artifact store, then launches a job in a Kubernetes cluster
    taken from your environment or specified via a passed-on kubectl config.

    Args:
        image: the Docker Image to be used for this ZenML pipeline
        job_prefix: a custom prefix for your Jobs in Kubernetes
            (default: 'zenml-')
        extra_labels: additional labels for your Jobs in Kubernetes
        extra_annotations: additional annotations for your Jobs in Kubernetes
        namespace: a custom Kubernetes namespace for this pipeline
            (default: 'default')
        image_pull_policy: Kubernetes image pull policy.
            One of ['Always', 'Never', 'IfNotPresent'].
            (default: 'IfNotPresent')
        kubernetes_config_path: Path to your Kubernetes cluster connection
        config.
            (default: '~/.kube/config'
    """

    def __init__(self,
                 image: Text = ZENML_BASE_IMAGE_NAME,
                 job_prefix: Text = 'zenml-',
                 extra_labels: Dict[Text, Any] = None,
                 extra_annotations: Dict[Text, Any] = None,
                 namespace: Text = None,
                 image_pull_policy: Text = ImagePullPolicy.IfNotPresent.name,
                 kubernetes_config_path: Text = DEFAULT_K8S_CONFIG,
                 **kwargs):
        self.image = image
        self.job_prefix = job_prefix
        self.extra_labels = extra_labels  # custom k8s labels
        self.extra_annotations = extra_annotations  # custom k8s annotations
        self.namespace = namespace
        self.image_pull_policy = image_pull_policy
        assert image_pull_policy in ImagePullPolicy.__members__.keys()
        self.kubernetes_config_path = kubernetes_config_path

        super().__init__(
            image=image,
            job_prefix=job_prefix,
            extra_labels=extra_labels,
            extra_annotations=extra_annotations,
            namespace=namespace,
            image_pull_policy=image_pull_policy,
            kubernetes_config_path=kubernetes_config_path,
            **kwargs,
        )

    def create_job_object(self, config):
        pipeline_name = config[keys.GlobalKeys.PIPELINE][
            keys.PipelineKeys.ARGS][keys.PipelineDetailKeys.NAME]
        job_name = to_dns1123(f'{self.job_prefix}{pipeline_name}', length=63)
        labels = self.extra_labels or {}
        job_labels = {
            "app": "zenml",
            "pipeline": pipeline_name,
            "datasource-id": config[keys.GlobalKeys.PIPELINE][
                keys.PipelineKeys.DATASOURCE][keys.DatasourceKeys.ID],
            "pipeline-id": get_id(pipeline_name)
        }
        labels.update(job_labels)  # make sure our labels are present

        config_encoded = base64.b64encode(json.dumps(config).encode()).decode(
            'utf-8')  # kubernetes needs the config as string

        command = ['python', '-m', K8S_ENTRYPOINT, 'run_pipeline',
                   '--config_b64', config_encoded]

        container = k8s_client.V1Container(
            name=job_name,
            image=self.image,
            command=command,
            image_pull_policy=self.image_pull_policy
        )

        # Create and configure a spec section
        template = k8s_client.V1PodTemplateSpec(
            metadata=k8s_client.V1ObjectMeta(labels=labels),
            spec=k8s_client.V1PodSpec(restart_policy="Never",
                                      containers=[container]))

        # Create the specification of deployment
        spec = k8s_client.V1JobSpec(
            template=template,
            backoff_limit=1)

        # Instantiate the job object
        job = k8s_client.V1Job(
            api_version="batch/v1",
            kind="Job",
            metadata=k8s_client.V1ObjectMeta(
                annotations=self.extra_annotations,
                labels=labels,
                name=job_name,
                namespace=self.namespace),
            spec=spec)

        return job

    def launch_job(self, config: Dict[Text, Any]):
        try:
            k8s_config.load_kube_config(self.kubernetes_config_path)
        except ConfigException as cfg_exc:
            logger.error("The path you provided does not contain a valid"
                         " Kubernetes config.")
            raise
        batch_client = k8s_client.BatchV1Api()
        job_object = self.create_job_object(config)

        namespace = self.namespace or 'default'

        api_response = batch_client.create_namespaced_job(
            body=job_object,
            namespace=namespace)
        logger.info(
            f'Created k8s {api_response.kind} '
            f'({api_response.api_version}): '
            f'{api_response.metadata.name}')

        return None

    def run(self, config: Dict[Text, Any]):
        # Extract the paths to create the tar
        logger.info('Orchestrating pipeline on Kubernetes..')

        repo: Repository = Repository.get_instance()
        repo_path = repo.path
        config_dir = repo.zenml_config.config_dir
        tar_file_name = \
            f'{EXTRACTED_TAR_DIR_NAME}_{str(int(time.time()))}.tar.gz'
        path_to_tar = os.path.join(config_dir, tar_file_name)

        # Create tarfile but exclude .zenml folder if exists
        path_utils.create_tarfile(repo_path, path_to_tar)
        logger.info(f'Created tar of current repository at: {path_to_tar}')

        # Upload tar to artifact store
        store_path = config[keys.GlobalKeys.ARTIFACT_STORE]
        store_staging_area = os.path.join(store_path, STAGING_AREA)
        store_path_to_tar = os.path.join(store_staging_area, tar_file_name)
        path_utils.copy(path_to_tar, store_path_to_tar)
        logger.info(f'Copied tar to artifact store at: {store_path_to_tar}')

        # Remove tar
        path_utils.rm_dir(path_to_tar)
        logger.info(f'Removed tar at: {path_to_tar}')

        # Append path of tar in config orchestrator utils
        config[keys.GlobalKeys.BACKEND][keys.BackendKeys.ARGS][
            TAR_PATH_ARG] = store_path_to_tar

        # Launch the instance
        self.launch_job(config)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import re
from typing import Text


def to_dns1123(name: Text, length=253):
    if len(name) > length:
        name = name[:length]
    name = re.sub(r'[^a-z0-9-\.]([^a-z0-9]{0,61}[^a-z0-9])?', r'-', name)
    return name.lower()

def get_id(text: Text):
    id_pattern = r'[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}'
    matches = re.findall(id_pattern, text)
    if not matches:
        match = None
    else:  # limit to the first result to prevent weirdness
        match = matches[0]
    return match




#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the identity split."""

from typing import Text, List

from zenml.steps.split import BaseSplitStep
from zenml.steps.split import constants


class NoSplit(BaseSplitStep):
    """
    No split function. Use this to pass your entire data forward completely
    unchanged.
    """

    def __init__(
            self,
            statistics=None,
            schema=None,
    ):
        """
        No split constructor.

        Use this to pass your entire data forward completely
        unchanged.

        Args:
            statistics: Parsed statistics from a preceding StatisticsGen.
            schema: Parsed schema from a preceding SchemaGen.
        """
        super().__init__(statistics=statistics, schema=schema)

    def partition_fn(self, element, n):
        """
        Function for no split on data; to be used in a beam.Partition.
        Args:
            element: Data point, given as a tf.train.Example.
            n: Number of splits, unused here.

        Returns:
            An integer n, where 0 ≤ n ≤ num_partitions - 1.
        """
        return 0

    def get_split_names(self) -> List[Text]:
        return [constants.NOSPLIT]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Postgres Datasource definition"""

from typing import Callable
from typing import Text, Dict

from beam_nuggets.io import relational_db

from zenml.datasources import BaseDatasource
from zenml.utils.beam_utils import WriteToTFRecord


class PostgresDatasource(BaseDatasource):
    """ZenML Postgres datasource definition.

    Use this for pipelines sourcing directly from a Postgres database table.
    """

    def __init__(
            self,
            name: Text,
            username: Text,
            password: Text,
            database: Text,
            table: Text,
            host: Text = 'localhost',
            port: int = 5432,
            query_limit: int = None,
            schema: Dict = None,
            **kwargs):
        """
        Initialize Postgres source. This creates a DataPipeline that
        essentially performs the following query using Apache Beam.

        `SELECT * FROM dataset.table LIMIT query_limit`
        Args:
            name: Name of datasource.
            username: Username of database user.
            password: Password to connect to database.
            database: Name of the target database.
            table: Name of the target table.
            host: Host of database.
            port: Port to connect to with database (default 5432)
            query_limit: Max number of rows to fetch.
            schema: Dict specifying schema.
        """
        self.username = username
        self.password = password
        self.database = database
        self.table = table
        self.host = host
        self.port = port
        self.query_limit = query_limit
        self.schema = schema

        super().__init__(
            name,
            username=self.username,
            password=self.password,
            database=self.database,
            table=self.table,
            host=self.host,
            port=self.port,
            query_limit=self.query_limit,
            schema=self.schema,
            **kwargs
        )

    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        query = f'SELECT * FROM {self.table}'

        if self.query_limit is not None:
            query += f'\nLIMIT {self.query_limit}'

        source_config = relational_db.SourceConfiguration(
            drivername='postgresql+pg8000',
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            database=self.database,
        )

        with make_beam_pipeline() as p:
            p | "Reading Postgres" >> relational_db.ReadFromDB(
                source_config=source_config,
                table_name=self.table,
                query=query) \
            | WriteToTFRecord(self.schema, output_path)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Orchestrator for simple GCP VM backend"""

import base64
import json
import os
import time
from typing import Dict, Any
from typing import Text

import googleapiclient.discovery
from google.oauth2 import service_account as sa

from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.repo import Repository
from zenml.standards import standard_keys as keys
from zenml.utils import path_utils
from zenml.utils import requirement_utils
from zenml.constants import ZENML_BASE_IMAGE_NAME, \
    ZENML_TRAINER_IMAGE_NAME, GCP_ENTRYPOINT
from zenml.logger import get_logger

requirement_utils.check_integration(requirement_utils.GCP_INTEGRATION)

logger = get_logger(__name__)

EXTRACTED_TAR_DIR_NAME = 'zenml_working'
TAR_PATH_ARG = 'tar_path'
STAGING_AREA = 'staging'


class OrchestratorGCPBackend(OrchestratorBaseBackend):
    """
    Orchestrates pipeline in a GCP Compute Instance.

    This orchestrator creates a .tar.gz of the current ZenML repository, sends
    it over to the artifact store, then launches a VM with the specified image.
    After pipeline is done, the VM automatically gets brought down, regardless
    of whether the pipeline failed or not. To see logs of the pipeline, use
    Logs Explorer <https://console.cloud.google.com/logs/> and filter for
    `logName="projects/<project_name>/logs/gcplogs-docker-driver"`. After
    running the VM, the logger returns the link directly to the logs.
    """

    def __init__(self,
                 project,
                 cloudsql_connection_name,
                 machine_type: Text = 'e2-medium',
                 gpu: Text = None,
                 gpu_count: int = 0,
                 zone: Text = 'europe-west1-b',
                 instance_name: Text = None,
                 disk_size: int = 100,
                 image: Text = None,
                 source_disk_image: Text = None,
                 preemptible: bool = True,
                 service_account: Text = None,
                 **kwargs):
        """
        Initialize a GCP VM to orchestrate a pipeline. Users have the option
        to run it with or without a GPU. In cases where a GPU is used,
        the `image` and `source_disk_image` are both adapted to be
        CUDA-compatible. This is convenient for the `TrainingPipeline`
        especially.

        Example:
            a) Without GPU:
            ```
            OrchestratorGCPBackend(
                project='my_project_id',
                cloudsql_connection_name='my_project_id:my_region:conn_name'
                machine_type='e2-medium'
            )
            ```
            In the above case, a smaller `image` is used for the
            orchestration of the pipeline, so the load-up time is faster. Use
            for smaller datasets or where a GPU is not required for speed-up
            training.

            b) With GPU:
            ```
            OrchestratorGCPBackend(
                project='my_project_id',
                cloudsql_connection_name='my_project_id:my_region:conn_name'
                machine_type='n1-standard-4',
                gpu='nvidia-tesla-k80',
            )
            ```
            Here, a large `image` is used for orchestration of the pipeline.
            The attached k80 GPU is leveraged for faster training. Note that
            not all machine_type are compatible with attached GPU! Make sure
            to check Google Cloud Platform documentation for a full list.

        Args:
            project: GCP project_id.
            cloudsql_connection_name: Cloud SQL instance name in the form
            {GCP_PROJECT}:{GCP_REGION}:{GCP_CLOUD_SQL_INSTANCE_NAME}

            gpu: (optional) GPU type to attach to VM. If gpu is specified,
            default `image` and `source_disk_image` are both modified. Full
            list of options [here](
            https://cloud.google.com/compute/docs/gpus/create-vm-with-gpus
            #gcloud_1)
            zone: The zone where VM is launched.
            instance_name: Name of the instance.
            disk_size: Size (in GB) of disk to be used.
            preemptible: Set True to use preemtible instance for reduced costs.
            image: The image in which the pipeline actually runs.
            machine_type: VM Machine type. Full list [here](
            https://cloud.google.com/compute/docs/machine-types)
            source_disk_image: The image of the underlying VM.
            service_account: Optional path to service account json file.
        """
        self.project = project
        self.cloudsql_connection_name = cloudsql_connection_name
        self.zone = zone
        self.image = image
        self.machine_type = machine_type
        self.preemptible = preemptible

        if instance_name is None:
            self.instance_name = 'zenml-' + str(int(time.time()))
        else:
            self.instance_name = instance_name

        self.gpu = gpu
        self.gpu_count = gpu_count

        if service_account:
            scopes = ['https://www.googleapis.com/auth/cloud-platform']
            self.credentials = \
                sa.Credentials.from_service_account_file(
                    service_account, scopes=scopes)
        else:
            self.credentials = None

        # Resolve images based on GPU
        if image is None:
            # use gpu image if a gpu is attached
            if self.gpu is None:
                self.image = ZENML_BASE_IMAGE_NAME
            else:
                self.image = ZENML_TRAINER_IMAGE_NAME

        if source_disk_image is None:
            compute = self._get_compute()
            if self.gpu is None:
                # get latest image from cos-85-lts family. As of Jan 28 2021
                # it is: cos-85-13310-1041-38
                image_response = compute.images().getFromFamily(
                    project='cos-cloud',
                    family='cos-85-lts').execute()
                source_disk_image = image_response['selfLink']
            else:
                # get latest image from common-dl family. As of Jan 28 2021
                # it is: 'c0-deeplearning-common-cu110-v20210121-debian-10'
                image_response = compute.images().getFromFamily(
                    project='deeplearning-platform-release',
                    family='common-cu110').execute()
                source_disk_image = image_response['selfLink']

        self.source_disk_image = source_disk_image
        self.disk_size = disk_size

        super().__init__(
            project=project,
            cloudsql_connection_name=cloudsql_connection_name,
            image=image,
            zone=zone,
            instance_name=instance_name,
            machine_type=machine_type,
            preemptible=preemptible,
            service_account=service_account,
            gpu=gpu,
            disk_size=disk_size,
            source_disk_image=source_disk_image,
            gpu_count=gpu_count,
            **kwargs,
        )

    def launch_instance(self, config: Dict[Text, Any]):
        """
        This function launches a GCP compute instance.

        Args:
            config: a ZenML config dict
        """
        # Instantiate google compute service

        # Configure the machine
        machine_type = f"zones/{self.zone}/machineTypes/{self.machine_type}"
        s_script_name = 'startup-script-gpu.sh' if self.gpu else \
            'startup-script.sh'
        startup_script = open(os.path.join(
            os.path.dirname(__file__), s_script_name), 'r').read()

        config_encoded = base64.b64encode(json.dumps(config).encode())
        c_params = f'python -m {GCP_ENTRYPOINT} run_pipeline --config_b64 ' \
                   f'{config_encoded}'

        compute_config = {
            "kind": "compute#instance",
            'name': self.instance_name,
            'zone': f'projects/{self.project}/zones/{self.zone}',
            'machineType': machine_type,
            "displayDevice": {
                "enableDisplay": False
            },
            # Specify if preemtible
            'scheduling': {'preemptible': self.preemptible},

            # Specify the boot disk and the image to use as a source.
            'disks': [
                {
                    "kind": "compute#attachedDisk",
                    'boot': True,
                    'autoDelete': True,
                    'initializeParams': {
                        'sourceImage': self.source_disk_image,
                        'diskSizeGb': str(self.disk_size)
                    }
                }
            ],

            # Specify a network interface with NAT to access the public
            # internet.
            'networkInterfaces': [{
                "kind": "compute#networkInterface",
                'network': 'global/networks/default',
                'accessConfigs': [
                    {"kind": "compute#accessConfig", 'type': 'ONE_TO_ONE_NAT',
                     'name': 'External NAT'}
                ]
            }],

            # Allow the instance to access cloud storage and logging.
            "serviceAccounts": [
                {
                    "email": "default",
                    "scopes": [
                        "https://www.googleapis.com/auth/cloud-platform",
                        "https://www.googleapis.com/auth/sqlservice.admin"
                    ]
                }
            ],

            # Metadata is readable from the instance and allows you to
            # pass configuration from deployment scripts to instances.
            'metadata': {
                "kind": "compute#metadata",
                'items': [
                    {
                        # Startup script is automatically executed by the
                        # instance upon startup.
                        'key': 'startup-script',
                        'value': startup_script
                    },
                    {
                        'key': 'image_name',
                        'value': self.image,
                    },
                    {
                        'key': 'container_params',
                        'value': c_params,
                    },
                    {
                        'key': 'mlmd_target',
                        'value': self.cloudsql_connection_name,
                    }
                ]
            }
        }

        if self.gpu:
            compute_config["guestAccelerators"] = [
                {
                    "acceleratorCount": self.gpu_count,
                    "acceleratorType":
                        f"projects/{self.project}/zones/{self.zone}"
                        f"/acceleratorTypes/{self.gpu}"
                }
            ]

        logger.info(
            f'Launching instance {self.instance_name} of type '
            f'{self.machine_type} in project: {self.project} in zone '
            f'{self.zone}')

        try:
            compute = self._get_compute()
            res = compute.instances().insert(
                project=self.project,
                zone=self.zone,
                body=compute_config).execute()
        except Exception as e:
            raise AssertionError(f"GCP VM failed to launch with the following "
                                 f"error: {str(e)}")

        logger.info(f'Launched instance {self.instance_name} with ID: '
                    f'{res["targetId"]}')
        log_link = \
            f'https://console.cloud.google.com/logs/query;query=logName%3D' \
            f'%22projects%2F{self.project}%2Flogs%2Fgcplogs-docker-driver%22' \
            f'%0Aresource.labels.instance_id%3D%22' \
            f'{res["targetId"]}%22?' \
            f'project={self.project}&folder=true&query=%0A'
        logger.info(f"View logs at: {log_link}")
        return res

    def run(self, config: Dict[Text, Any]):
        """
        This run function essentially calls an underlying TFX orchestrator run.
        However it is meant as a higher level abstraction with some
        opinionated decisions taken.

        Args:
            config: a ZenML config dict
        """
        # Extract the paths to create the tar
        logger.info('Orchestrating pipeline on GCP..')

        repo: Repository = Repository.get_instance()
        repo_path = repo.path
        config_dir = repo.zenml_config.config_dir
        tar_file_name = \
            f'{EXTRACTED_TAR_DIR_NAME}_{str(int(time.time()))}.tar.gz'
        path_to_tar = os.path.join(config_dir, tar_file_name)

        # Create tarfile but exclude .zenml folder if exists
        path_utils.create_tarfile(repo_path, path_to_tar)
        logger.info(f'Created tar of current repository at: {path_to_tar}')

        # Upload tar to artifact store
        store_path = config[keys.GlobalKeys.ARTIFACT_STORE]
        store_staging_area = os.path.join(store_path, STAGING_AREA)
        store_path_to_tar = os.path.join(store_staging_area, tar_file_name)
        path_utils.copy(path_to_tar, store_path_to_tar)
        logger.info(f'Copied tar to artifact store at: {store_path_to_tar}')

        # Remove tar
        path_utils.rm_dir(path_to_tar)
        logger.info(f'Removed tar at: {path_to_tar}')

        # Append path of tar in config orchestrator utils
        config[keys.GlobalKeys.BACKEND][keys.BackendKeys.ARGS][
            TAR_PATH_ARG] = store_path_to_tar

        # Launch the instance
        self.launch_instance(config)

    def _get_compute(self):
        return googleapiclient.discovery.build(
            'compute', 'v1', credentials=self.credentials)



import json
import os

import requests

CORTEX_ENDPOINT = os.getenv('CORTEX_ENV_ENDPOINT')
CORTEX_MODEL_NAME = os.getenv('CORTEX_MODEL_NAME')
assert CORTEX_ENDPOINT
assert CORTEX_MODEL_NAME

ENDPOINT = f'{CORTEX_ENDPOINT}/{CORTEX_MODEL_NAME}'


def make_predict(json_request):
    headers = {
        'Content-Type': 'application/json',
    }
    data = json.dumps(json_request)
    response = requests.post(ENDPOINT, headers=headers, data=data)
    return response.json()


def main():
    data = {
        "times_pregnant": 1,
        "pgc": 148,
        "dbp": 72,
        "tst": 35,
        "insulin": 100,
        "bmi": 33.6,
        "pedigree": 0.627,
        "age": 50,
    }
    response = make_predict(data)
    print(
        f'The probability of this person to have diabetes: '
        f'{response["has_diabetes"][0]}')


if __name__ == '__main__':
    main()



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.steps.sequencer.standard_sequencer.methods import \
    methods_resampling, methods_filling
from zenml.utils.preprocessing_utils import MethodDescriptions


class ResamplingMethods(MethodDescriptions):
    MODES = {'mean': (methods_resampling.resample_mean, []),
             'threshold': (methods_resampling.resample_thresholding,
                           ['cond', 'c_value', 'threshold', 'set_value']),
             'mode': (methods_resampling.resample_mode, []),
             'median': (methods_resampling.resample_median, [])}


class FillingMethods(MethodDescriptions):
    MODES = {'forward': (methods_filling.forward_f, []),
             'backwards': (methods_filling.backwards_f, []),
             'min': (methods_filling.min_f, []),
             'max': (methods_filling.max_f, []),
             'mean': (methods_filling.mean_f, []),
             'custom': (methods_filling.custom_f, ['custom_value'])}



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


import apache_beam as beam
import tensorflow as tf


class ConvertToDataframe(beam.DoFn):
    """
    Beam PTransform responsible for converting the incoming Arrow table into
    a pandas dataframe
    """

    def process(self, element, *args, **kwargs):
        return [element.to_pandas()]


def df_to_example(instance) -> tf.train.Example:
    """
    Auxiliary function to create a tf.train.Example from a pandas dataframe

    :param instance: pd.DataFrame, the datapoint
    :return: tf.train.Example, the result
    """
    d = {col: instance[col] for col in instance.columns}

    feature = {}
    for key, value in d.items():
        if value is None:
            raise Exception('The value can not possibly be None!')
        elif value.dtype == int:
            data = value.tolist()
            feature[key] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=data))
        elif value.dtype == float:
            data = value.tolist()
            feature[key] = tf.train.Feature(
                float_list=tf.train.FloatList(value=data))
        else:
            data = list(map(tf.compat.as_bytes, value.tolist()))
            feature[key] = tf.train.Feature(
                bytes_list=tf.train.BytesList(value=data))
    return tf.train.Example(features=tf.train.Features(feature=feature))


def serialize(instance):
    """
    Helper function to serialize a tf.train.Example in the beam pipeline

    :param instance: the input, tf.train.Example
    :return: serialized version of the input
    """
    return instance.SerializeToString(deterministic=True)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from tfx.orchestration import data_types
from tfx.orchestration import metadata
from tfx.orchestration import pipeline
from tfx.orchestration.config import config_utils
from tfx.orchestration.local.local_dag_runner import LocalDagRunner

from zenml.logger import get_logger

logger = get_logger(__name__)


class ZenMLLocalDagRunner(LocalDagRunner):
    """
    This is the almost the same as the super class from tfx:
    tfx.orchestration.base.local_dag_runner.LocalDagRunner with the exception
    being that the pipeline_run is not overridden. Full credit to Google LLC
    for the original source code found at:
    https://github.com/tensorflow/tfx/blob/master/tfx/orchestration/local/
    """

    def run(self, tfx_pipeline: pipeline.Pipeline) -> None:
        for component in tfx_pipeline.components:
            (component_launcher_class, component_config) = (
                config_utils.find_component_launch_info(self._config,
                                                        component))
            driver_args = data_types.DriverArgs(
                enable_cache=tfx_pipeline.enable_cache)
            metadata_connection = metadata.Metadata(
                tfx_pipeline.metadata_connection_config)
            component_launcher = component_launcher_class.create(
                component=component,
                pipeline_info=tfx_pipeline.pipeline_info,
                driver_args=driver_args,
                metadata_connection=metadata_connection,
                beam_pipeline_args=tfx_pipeline.beam_pipeline_args,
                additional_pipeline_args=tfx_pipeline
                    .additional_pipeline_args,
                component_config=component_config)
            logger.info('Component %s is running.', component.id)
            component_launcher.launch()
            logger.info('Component %s is finished.', component.id)



import os

from zenml.backends.orchestrator import OrchestratorGCPBackend
from zenml.datasources import CSVDatasource
from zenml.metadata import MySQLMetadataStore
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository, ArtifactStore
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.exceptions import AlreadyExistsException

from zenml.utils.naming_utils import transformed_label_name

GCP_PROJECT = os.getenv('GCP_PROJECT')
GCP_BUCKET = os.getenv('GCP_BUCKET')
GCP_REGION = os.getenv('GCP_REGION')
MYSQL_DB = os.getenv('MYSQL_DB')
MYSQL_USER = os.getenv('MYSQL_USER')
MYSQL_PWD = os.getenv('MYSQL_PWD')
MYSQL_HOST = os.getenv('MYSQL_HOST', '127.0.0.1')
MYSQL_PORT = os.getenv('MYSQL_PORT', 3306)
GCP_CLOUD_SQL_INSTANCE_NAME = os.getenv('GCP_CLOUD_SQL_INSTANCE_NAME',
                                        MYSQL_DB)
CONNECTION_NAME = f'{GCP_PROJECT}:{GCP_REGION}:{GCP_CLOUD_SQL_INSTANCE_NAME}'

assert GCP_BUCKET
assert GCP_PROJECT
assert GCP_REGION
assert MYSQL_DB
assert MYSQL_USER
assert MYSQL_PWD

# Run the pipeline on a Google Cloud VM.
# The metadata store and artifact store should be accessible by the
# orchestrator VM.

# Define the metadata store
metadata_store = MySQLMetadataStore(
    host=MYSQL_HOST,
    port=int(MYSQL_PORT),
    database=MYSQL_DB,
    username=MYSQL_USER,
    password=MYSQL_PWD,
)

# Define the artifact store
artifact_store = ArtifactStore(
    os.path.join(GCP_BUCKET, 'gcp_orchestrated/artifact_store'))

# Define the orchestrator backend
orchestrator_backend = OrchestratorGCPBackend(
    cloudsql_connection_name=CONNECTION_NAME,
    project=GCP_PROJECT,
    preemptible=True,  # reduce costs by using preemptible instances
)


# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv',
                       backend=orchestrator_backend,
                       metadata_store=metadata_store,
                       artifact_store=artifact_store)
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

ds.commit()

# # Add a split
# training_pipeline.add_split(RandomSplit(
#     split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))
#
# # Add a preprocessing unit
# training_pipeline.add_preprocesser(
#     StandardPreprocesser(
#         features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
#                   'pedigree', 'age'],
#         labels=['has_diabetes'],
#         overwrite={'has_diabetes': {
#             'transform': [{'method': 'no_transform', 'parameters': {}}]}}
#     ))
#
# # Add a trainer
# training_pipeline.add_trainer(TFFeedForwardTrainer(
#     loss='binary_crossentropy',
#     last_activation='sigmoid',
#     output_units=1,
#     metrics=['accuracy'],
#     epochs=20))
#
# # Add an evaluator
# training_pipeline.add_evaluator(
#     TFMAEvaluator(slices=[['has_diabetes']],
#                   metrics={transformed_label_name('has_diabetes'):
#                      ['binary_crossentropy', 'binary_accuracy']}))
#
#
# # Run the pipeline
# training_pipeline.run(
#     backend=orchestrator_backend,
#     metadata_store=metadata_store,
#     artifact_store=artifact_store,
# )



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the GCAIP Training Executor"""

from tfx.extensions.google_cloud_ai_platform.trainer.executor import \
    GenericExecutor

from zenml.components.trainer.executor import ZenMLTrainerExecutor

# Keys to the items in custom_config passed as a part of exec_properties.
TRAINING_ARGS_KEY = 'ai_platform_training_args'
JOB_ID_KEY = 'ai_platform_training_job_id'
_CUSTOM_CONFIG_KEY = 'custom_config'


class ZenMLTrainerGCAIPExecutor(GenericExecutor):
    """Start a trainer job on Google Cloud AI Platform using a default
    Trainer."""

    def _GetExecutorClass(self):
        return ZenMLTrainerExecutor



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Dict, Text, Any, List
from typing import Optional

import tensorflow as tf
from tfx.components.common_nodes.importer_node import ImporterNode
from tfx.components.schema_gen.component import SchemaGen
from tfx.components.statistics_gen.component import StatisticsGen
from tfx.types import standard_artifacts

from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.components import BulkInferrer
from zenml.datasources import BaseDatasource
from zenml.enums import GDPComponent
from zenml.logger import get_logger
from zenml.metadata import ZenMLMetadataStore
from zenml.pipelines import BasePipeline
from zenml.repo import ArtifactStore
from zenml.standards import standard_keys as keys
from zenml.standards.standard_keys import StepKeys
from zenml.steps import BaseStep
from zenml.steps.inferrer import BaseInferrer
from zenml.utils import path_utils
from zenml.utils.post_training.post_training_utils import \
    convert_raw_dataset_to_pandas
from zenml.utils.post_training.post_training_utils import \
    view_statistics, view_schema, get_feature_spec_from_schema

logger = get_logger(__name__)


class BatchInferencePipeline(BasePipeline):
    """BatchInferencePipeline definition to run batch inference pipelines.

    A BatchInferencePipeline is used to run an inference based on a
    TrainingPipeline.
    """
    PIPELINE_TYPE = 'infer'

    def __init__(self,
                 model_uri: Text,
                 name: Text = None,
                 enable_cache: Optional[bool] = True,
                 steps_dict: Dict[Text, BaseStep] = None,
                 backend: OrchestratorBaseBackend = None,
                 metadata_store: Optional[ZenMLMetadataStore] = None,
                 artifact_store: Optional[ArtifactStore] = None,
                 datasource: Optional[BaseDatasource] = None,
                 pipeline_name: Optional[Text] = None):
        """
        Construct a Batch Inference pipeline. This is a pipeline that allows
        for offline batch inference.

        Args:
            name: Outward-facing name of the pipeline.
            model_uri: URI for a model, usually generated by
            TrainingPipeline and retrieved by
            `training_pipeline.get_model_uri()`.
            labels: List of labels to be predicted by the model.
            pipeline_name: A unique name that identifies the pipeline after
             it is run.
            enable_cache: Boolean, indicates whether or not caching
             should be used.
            steps_dict: Optional dict of steps.
            backend: Orchestrator backend.
            metadata_store: Configured metadata store. If None,
             the default metadata store is used.
            artifact_store: Configured artifact store. If None,
             the default artifact store is used.
        """
        if model_uri is None:
            raise AssertionError('model_uri cannot be None.')
        self.model_uri = model_uri
        super(BatchInferencePipeline, self).__init__(
            name=name,
            enable_cache=enable_cache,
            steps_dict=steps_dict,
            backend=backend,
            metadata_store=metadata_store,
            artifact_store=artifact_store,
            datasource=datasource,
            pipeline_name=pipeline_name,
            model_uri=model_uri,
        )

    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Creates an inference pipeline out of TFX components.

        A inference pipeline is used to run a batch of data through a
        ML model via the BulkInferrer TFX component.

        Args:
            config: Dict. Contains a ZenML configuration used to build the
             data pipeline.

        Returns:
            A list of TFX components making up the data pipeline.
        """
        component_list = []

        # if there are no commits, then lets make one
        if self.datasource.is_empty:
            logger.info(
                f'Datasource {self.datasource.name} has no commits. Creating '
                f'the first one..')
            self.datasource_commit_id = self.datasource.commit()

        data_pipeline = self.datasource.get_data_pipeline_from_commit(
            self.datasource_commit_id)

        data = ImporterNode(
            instance_name=GDPComponent.DataGen.name,
            source_uri=data_pipeline.get_artifacts_uri_by_component(
                GDPComponent.DataGen.name)[0],
            artifact_type=standard_artifacts.Examples)
        component_list.extend([data])

        # Handle timeseries
        # TODO: [LOW] Handle timeseries
        # if GlobalKeys. in train_config:
        #     schema = ImporterNode(instance_name='Schema',
        #                           source_uri=spec['schema_uri'],
        #                           artifact_type=standard_artifacts.Schema)
        #
        #     sequence_transform = SequenceTransform(
        #         examples=data.outputs.examples,
        #         schema=schema,
        #         config=train_config,
        #         instance_name=GDPComponent.SequenceTransform.name)
        #     datapoints = sequence_transform.outputs.output
        #     component_list.extend([schema, sequence_transform])

        # Load from model_uri
        model = ImporterNode(
            instance_name=GDPComponent.Trainer.name,
            source_uri=self.model_uri,
            artifact_type=standard_artifacts.Model)

        model_result = model.outputs.result

        infer_cfg = config[keys.GlobalKeys.PIPELINE][keys.PipelineKeys.STEPS][
            keys.InferSteps.INFER]

        bulk_inferrer = BulkInferrer(
            source=infer_cfg[StepKeys.SOURCE],
            source_args=infer_cfg[StepKeys.ARGS],
            model=model_result,
            examples=data.outputs.result,
            instance_name=GDPComponent.Inferrer.name)

        statistics = StatisticsGen(
            examples=bulk_inferrer.outputs.predictions
        ).with_id(GDPComponent.DataStatistics.name)

        schema = SchemaGen(
            statistics=statistics.outputs.output,
        ).with_id(GDPComponent.DataSchema.name)

        component_list.extend([model, bulk_inferrer, statistics, schema])

        return component_list

    def add_infer_step(self, infer_step: BaseInferrer):
        self.steps_dict[keys.InferSteps.INFER] = infer_step

    def steps_completed(self) -> bool:
        mandatory_steps = [keys.InferSteps.DATA, keys.InferSteps.INFER]
        for step_name in mandatory_steps:
            if step_name not in self.steps_dict.keys():
                raise AssertionError(
                    f'Mandatory step {step_name} not added.')
        return True

    def view_statistics(self, magic: bool = False, port: int = 0):
        """
        View statistics for infer pipeline in HTML.

        Args:
            magic (bool): Creates HTML page if False, else
            creates a notebook cell.
            port (int): Port at which to launch the statistics facet.
        """
        uri = self.get_artifacts_uri_by_component(
            GDPComponent.DataStatistics.name)[0]
        view_statistics(uri, magic, port)

    def view_schema(self):
        """View schema of data flowing in pipeline."""
        uri = self.get_artifacts_uri_by_component(
            GDPComponent.DataSchema.name)[0]
        view_schema(uri)

    def get_predictions(self, sample_size: int = 100000):
        """
        Samples prediction data as a pandas DataFrame.

        Args:
            sample_size: # of rows to sample.
        """
        base_uri = self.get_artifacts_uri_by_component(
            GDPComponent.Inferrer.name)[0]
        data_files = path_utils.list_dir(os.path.join(base_uri, 'examples'))
        dataset = tf.data.TFRecordDataset(data_files, compression_type='GZIP')
        schema_uri = self.get_artifacts_uri_by_component(
            GDPComponent.DataSchema.name)[0]
        spec = get_feature_spec_from_schema(schema_uri)
        return convert_raw_dataset_to_pandas(dataset, spec, sample_size)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Any, Dict, List, Text

import apache_beam as beam
import tensorflow as tf
from tfx import types
from tfx.dsl.components.base.base_executor import BaseExecutor
from tfx.types import artifact_utils
from tfx.types.artifact_utils import get_split_uri
from tfx.utils import io_utils

from zenml.components.data_gen.constants import DATA_SPLIT_NAME
from zenml.components.split_gen import constants
from zenml.components.split_gen.utils import parse_schema, \
    parse_statistics
from zenml.standards.standard_keys import StepKeys
from zenml.steps.split import BaseSplitStep
from zenml.steps.split.constants import SKIP
from zenml.utils import source_utils
from zenml.logger import get_logger

logger = get_logger(__name__)


@beam.ptransform_fn
@beam.typehints.with_input_types(bytes)
@beam.typehints.with_output_types(beam.pvalue.PDone)
def WriteSplit(example_split: beam.pvalue.PCollection,
               output_split_path: Text) -> beam.pvalue.PDone:
    """Shuffles and writes output split."""
    out_dir = os.path.join(output_split_path, 'data_tfrecord')
    logger.debug('Writing output TFRecords to {}'.format(
        out_dir
    ))
    return (example_split
            | 'Write' >> beam.io.WriteToTFRecord(
                out_dir,
                file_name_suffix='.gz'))


class Executor(BaseExecutor):
    def Do(self, input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:
        """
        Write description regarding this beautiful executor.

        Args:
            input_dict:
            output_dict:
            exec_properties:
        """
        self._log_startup(input_dict, output_dict, exec_properties)

        schema = parse_schema(input_dict=input_dict)

        statistics = parse_statistics(
            split_name=DATA_SPLIT_NAME,
            statistics=input_dict[constants.STATISTICS])

        source = exec_properties[StepKeys.SOURCE]
        args = exec_properties[StepKeys.ARGS]

        # pass the schema and stats straight to the Step
        args[constants.SCHEMA] = schema
        args[constants.STATISTICS] = statistics

        c = source_utils.load_source_path_class(source)
        split_step: BaseSplitStep = c(**args)

        # infer the names of the splits from the config
        split_names = split_step.get_split_names()

        # Get output split path
        examples_artifact = artifact_utils.get_single_instance(
            output_dict[constants.OUTPUT_EXAMPLES])
        if SKIP in split_names:
            sanitized_names = [name for name in split_names if name != SKIP]
            examples_artifact.split_names = artifact_utils.encode_split_names(
                sanitized_names)
        else:
            examples_artifact.split_names = artifact_utils.encode_split_names(
                split_names)

        split_uris = []
        for artifact in input_dict[constants.INPUT_EXAMPLES]:
            for split in artifact_utils.decode_split_names(
                    artifact.split_names):
                uri = os.path.join(artifact.uri, split)
                split_uris.append((split, uri))

        with self._make_beam_pipeline() as p:
            # The outer loop will for now only run once
            for split, uri in split_uris:
                input_uri = io_utils.all_files_pattern(uri)

                new_splits = (
                        p
                        | 'ReadData.' + split >> beam.io.ReadFromTFRecord(file_pattern=input_uri)
                        | beam.Map(tf.train.Example.FromString)
                        | 'Split' >> beam.Partition(split_step.partition_fn,
                                                    split_step.get_num_splits()))

                for split_name, new_split in zip(split_names,
                                                 list(new_splits)):
                    if split_name != SKIP:
                        # WriteSplit function writes to TFRecord again
                        (new_split
                         | 'Serialize.' + split_name >> beam.Map(
                                    lambda x: x.SerializeToString())
                         | 'WriteSplit_' + split_name >> WriteSplit(
                                    get_split_uri(
                                        output_dict[constants.OUTPUT_EXAMPLES],
                                        split_name)))



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

# pylint disable=protected-access

import os
import shutil
from pathlib import Path

import pytest

import zenml
from zenml.backends import BaseBackend
from zenml.datasources import BaseDatasource
from zenml.metadata import ZenMLMetadataStore
from zenml.pipelines import BasePipeline
from zenml.repo import Repository, ZenMLConfig
from zenml.steps import BaseStep
from zenml.utils import path_utils
from zenml.logger import get_logger

logger = get_logger(__name__)

# Nicholas a way to get to the root
ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")

pipeline_root = os.path.join(TEST_ROOT, "pipelines")


@pytest.fixture
def repo():
    return Repository.get_instance()


@pytest.fixture
def cleanup_metadata_store():
    def wrapper():
        # Remove metadata db after a test to avoid test failures by duplicated
        # data sources
        metadata_db_location = os.path.join(".zenml", "local_store",
                                            "metadata.db")
        try:
            os.remove(os.path.join(ZENML_ROOT, metadata_db_location))
        except Exception as e:
            logger.error(e)

    return wrapper


@pytest.fixture
def cleanup_pipelines_dir():
    def wrapper():
        repo: Repository = Repository.get_instance()
        pipelines_dir = repo.zenml_config.get_pipelines_dir()
        for p_config in path_utils.list_dir(pipelines_dir):
            try:
                os.remove(p_config)
            except Exception as e:
                logger.error(e)

    return wrapper


@pytest.fixture
def cleanup_artifacts():
    def wrapper():
        local_store_path = os.path.join(".zenml", "local_store")
        try:
            local_store = os.path.join(ZENML_ROOT, local_store_path)
            shutil.rmtree(local_store, ignore_errors=False)
        except Exception as e:
            logger.error(e)

    return wrapper


@pytest.fixture
def cleanup(cleanup_metadata_store, cleanup_artifacts):
    cleanup_metadata_store()
    cleanup_artifacts()


@pytest.fixture
def delete_config():
    def wrapper(filename):
        cfg = os.path.join(pipeline_root, filename)
        path_utils.rm_file(cfg)

    return wrapper


@pytest.fixture
def equal_backends():
    def wrapper(bck1: BaseBackend, bck2: BaseBackend, loaded=True):
        # There can be a "None" backend in a pipeline
        if bck1 is None and bck2 is None:
            return True
        if sum(b is None for b in [bck1, bck2]) == 1:
            return False

        equal = False
        equal |= bck1._kwargs == bck2._kwargs
        equal |= bck1._source == bck2._source
        if loaded:
            equal |= bck1._immutable != bck2._immutable
        else:
            equal |= bck1._immutable == bck2._immutable

        return equal

    return wrapper


@pytest.fixture
def equal_steps(equal_backends):
    def wrapper(step1: BaseStep, step2: BaseStep, loaded=True):
        # There can be a "None" step e.g. in get_data_step()
        # in a BaseDatasource
        if step1 is None and step2 is None:
            return True
        if sum(s is None for s in [step1, step2]) == 1:
            return False

        equal = False
        equal |= step1._kwargs == step2._kwargs
        equal |= equal_backends(step1.backend, step2.backend, loaded=loaded)
        equal |= step1._source == step2._source
        if loaded:
            equal |= step1._immutable != step2._immutable
        else:
            equal |= step1._immutable == step2._immutable

        return equal

    return wrapper


@pytest.fixture
def equal_datasources(equal_steps):
    def wrapper(ds1: BaseDatasource, ds2: BaseDatasource, loaded=True):
        # There can be a "None" datasource in a pipeline
        if ds1 is None and ds2 is None:
            return True
        if sum(d is None for d in [ds1, ds2]) == 1:
            return False

        equal = False
        equal |= ds1.name == ds2.name
        equal |= ds1._id == ds2._id
        equal |= ds1._source == ds2._source
        equal |= ds1._source_args == ds2._source_args

        # TODO[LOW]: Add more checks for constructor kwargs, __dict__ etc.
        if loaded:
            equal |= ds1._immutable != ds2._immutable
        else:
            equal |= ds1._immutable == ds2._immutable

        return equal

    return wrapper


@pytest.fixture
def equal_pipelines(equal_backends, equal_steps, equal_datasources):
    def wrapper(p1: BasePipeline, p2: BasePipeline, loaded=True):
        # There can be a "None" datasource in a pipeline
        if p1 is None and p2 is None:
            return True
        if sum(p is None for p in [p1, p2]) == 1:
            return False

        equal = False
        equal |= p1.name == p2.name
        equal |= p1.PIPELINE_TYPE == p2.PIPELINE_TYPE
        equal |= p1.pipeline_name == p2.pipeline_name
        equal |= p1.enable_cache == p2.enable_cache
        equal |= p1._source == p2._source
        equal |= equal_backends(p1.backend, p2.backend, loaded=loaded)
        equal |= equal_datasources(p1.datasource, p2.datasource, loaded=loaded)
        if loaded:
            equal |= p1._immutable != p2._immutable
        else:
            equal |= p1._immutable == p2._immutable
        try:
            for name, step in p1.steps_dict.items():
                p2_step = p2.steps_dict[name]
                equal |= equal_steps(step, p2_step, loaded=loaded)
        except KeyError:
            return False

        return equal

    return wrapper


@pytest.fixture
def equal_md_stores():
    def wrapper(md1: ZenMLMetadataStore, md2: ZenMLMetadataStore):
        # There can be a "None" datasource in a pipeline
        if md1 is None and md2 is None:
            return True
        if sum(d is None for d in [md1, md2]) == 1:
            return False
        equal = False
        equal |= md1.__dict__ == md2.__dict__
        equal |= md1.STORE_TYPE == md2.STORE_TYPE

        return equal

    return wrapper


@pytest.fixture
def equal_zenml_configs(equal_md_stores):
    def wrapper(cfg1: ZenMLConfig, cfg2: ZenMLConfig, loaded=True):
        # There can be a "None" datasource in a pipeline
        if cfg1 is None and cfg2 is None:
            return True
        if sum(d is None for d in [cfg1, cfg2]) == 1:
            return False

        # TODO[LOW]: Expand logic
        equal = False
        equal |= equal_md_stores(cfg1.metadata_store, cfg2.metadata_store)
        equal |= cfg1.pipelines_dir == cfg2.pipelines_dir
        equal |= cfg1.artifact_store.path == cfg2.artifact_store.path

        return equal

    return wrapper



from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import TrainingPipeline, BatchInferencePipeline
from zenml.repo import Repository
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.inferrer import TensorflowInferrer
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.utils.naming_utils import transformed_label_name

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))

# Add a trainer
training_pipeline.add_trainer(TFFeedForwardTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    output_units=1,
    metrics=['accuracy'],
    epochs=20))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(slices=[['has_diabetes']],
                  metrics={transformed_label_name('has_diabetes'): [
                      'binary_crossentropy',
                      'binary_accuracy']}))

# Run the pipeline locally
training_pipeline.run()

# Run inference
model_uri = training_pipeline.get_model_uri()
infer_pipeline = BatchInferencePipeline(
    model_uri=model_uri,
)
infer_pipeline.add_datasource(ds)
infer_pipeline.add_infer_step(
    TensorflowInferrer(
        labels=[transformed_label_name('has_diabetes')])
)
infer_pipeline.run()
df = infer_pipeline.get_predictions()
print(df.head())



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from abc import abstractmethod
from typing import Text, Dict

from ml_metadata.metadata_store import metadata_store

from zenml.enums import MLMetadataTypes
from zenml.enums import PipelineStatusTypes
from zenml.exceptions import DoesNotExistException
from zenml.logger import get_logger
from zenml.standards.standard_keys import MLMetadataKeys
from zenml.utils.print_utils import to_pretty_string, PrintStyles

logger = get_logger(__name__)


class ZenMLMetadataStore:
    STORE_TYPE = None
    RUN_TYPE_PROPERTY_NAME = 'run'

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @property
    def store(self):
        return metadata_store.MetadataStore(self.get_tfx_metadata_config())

    @classmethod
    def from_config(cls, config: Dict):
        """
        Converts ZenML config to ZenML Metadata Store.

        Args:
            config (dict): ZenML config block for metadata.
        """
        from zenml.metadata.metadata_wrapper_factory import \
            wrapper_factory
        store_type = config[MLMetadataKeys.TYPE]
        store_types = list(MLMetadataTypes.__members__.keys())
        if store_type not in store_types:
            raise AssertionError(f'store_type must be one of: {store_types}')
        args = config[MLMetadataKeys.ARGS]
        class_ = wrapper_factory.get_single_metadata_wrapper(store_type)

        try:
            obj = class_(**args)
        except TypeError as e:
            logger.error(str(e))
            import inspect
            args = inspect.getfullargspec(class_).args
            args.remove('self')
            raise Exception(f'args must include only: {args}')
        return obj

    @abstractmethod
    def get_tfx_metadata_config(self):
        pass

    def to_config(self) -> Dict:
        """
        Converts from ZenML Metadata store back to config.
        """
        store_types = list(MLMetadataTypes.__members__.keys())
        if self.STORE_TYPE not in store_types:
            raise AssertionError(f'store_type must be one of: {store_types}')

        args_dict = self.__dict__.copy()
        return {
            MLMetadataKeys.TYPE: self.STORE_TYPE,
            MLMetadataKeys.ARGS: args_dict
        }

    def get_data_pipeline_names_from_datasource_name(self,
                                                     datasource_name: Text):
        """
        Gets all data pipeline names from the datasource name.

        Args:
            datasource_name: name of datasource.
        """
        from zenml.pipelines.data_pipeline import DataPipeline

        run_contexts = self.store.get_contexts_by_type(
            self.RUN_TYPE_PROPERTY_NAME)

        # get the data pipelines only by doing an ugly hack. These data
        # pipelines will always start with `data_` and have no component
        # name at the end. this needs to change soon
        run_contexts = [x for x in run_contexts if
                        x.name.split('.')[1].startswith(
                            DataPipeline.PIPELINE_TYPE) and len(
                            x.name.split('.')) == 2]

        # now filter to the datasource name through executions
        pipelines_names = []
        for c in run_contexts:
            es = self.store.get_executions_by_context(c.id)
            for e in es:
                if 'name' in e.properties and e.properties[
                    'name'].string_value == datasource_name:
                    pipelines_names.append(c.name.split('.')[1])
        return pipelines_names

    def get_pipeline_status(self, pipeline) -> Text:
        """
        Query metadata store to find status of pipeline.

        Args:
            pipeline (BasePipeline): a ZenML pipeline object
        """
        try:
            components_status = self.get_components_status(pipeline)
        except:
            return PipelineStatusTypes.NotStarted.name

        for status in components_status.values():
            if status != 'complete' and status != 'cached':
                return PipelineStatusTypes.Running.name
        return PipelineStatusTypes.Succeeded.name

    def get_pipeline_executions(self, pipeline):
        """
        Get executions of pipeline.

        Args:
            pipeline (BasePipeline): a ZenML pipeline object
        """
        c = self.get_pipeline_context(pipeline)
        return self.store.get_executions_by_context(c.id)

    def get_components_status(self, pipeline):
        """
        Returns status of components in pipeline.

        Args:
            pipeline (BasePipeline): a ZenML pipeline object

        Returns: dict of type { component_name : component_status }
        """
        pipeline_executions = self.get_pipeline_executions(pipeline)
        return {
            e.properties['component_id'].string_value: e.properties[
                'state'].string_value for e in
            pipeline_executions
        }

    def get_artifacts_by_component(self, pipeline, component_name: Text):
        """
        Args:
            pipeline (BasePipeline): a ZenML pipeline object
            component_name:
        """
        # First , you get the execution associated with the component
        e = self.get_component_execution(pipeline, component_name)

        if e is None:
            raise DoesNotExistException(
                name=component_name,
                reason=f'The pipeline {pipeline.name} does not have the '
                       f'associated {component_name} Step.')

        # Second, you will get artifacts
        return self.get_artifacts_by_execution(e.id)

    def get_component_execution(self, pipeline, component_name: Text):
        pipeline_executions = self.get_pipeline_executions(pipeline)
        for e in pipeline_executions:
            # TODO: [LOW] Create a more refined way to find components.
            if component_name == e.properties['component_id'].string_value:
                return e

    def get_pipeline_context(self, pipeline):
        # We rebuild context for ml metadata here.
        prefix = pipeline.artifact_store.unique_id
        run_id = f'{prefix}.{pipeline.pipeline_name}'
        logger.debug(f'Looking for run_id {run_id} in metadata store: '
                     f'{self.to_config()}')
        run_context = self.store.get_context_by_type_and_name(
            type_name=self.RUN_TYPE_PROPERTY_NAME,
            context_name=run_id
        )
        if run_context is None:
            raise DoesNotExistException(
                name=pipeline.pipeline_name,
                reason=f'The pipeline does not exist in metadata store '
                       f'because it has not been run yet. Please run the '
                       f'pipeline before trying to fetch artifacts.')
        return run_context

    def get_artifacts_by_execution(self, execution_id):
        # First get all the events for this execution
        """
        Args:
            execution_id:
        """
        events = self.store.get_events_by_execution_ids([execution_id])

        artifact_ids = []
        for e in events:
            # MAJOR Assumptions: 4 means its output channel
            if e.type == 4:
                artifact_ids.append(e.artifact_id)
                break

        if not artifact_ids:
            raise AssertionError("This execution has no output artifact.")

        # Get artifacts
        return self.store.get_artifacts_by_id(artifact_ids)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file is used to create the package we'll publish to PyPI.

.. currentmodule:: setup.py
.. moduleauthor:: maiot GmbH <support@maiot.io>
"""

import importlib.util
import os
from codecs import open  # Use a consistent encoding.
from os import path
from pathlib import Path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the base version from the library.  (We'll find it in the `version.py`
# file in the src directory, but we'll bypass actually loading up the library.)
vspec = importlib.util.spec_from_file_location(
    "version",
    str(Path(__file__).resolve().parent /
        'zenml' / 'version.py')
)
vmod = importlib.util.module_from_spec(vspec)
vspec.loader.exec_module(vmod)
version = getattr(vmod, '__version__')

# Same method for the requirements
vspec = importlib.util.spec_from_file_location(
    "requirement_utils",
    str(Path(__file__).resolve().parent
        / 'zenml' / 'utils' / "requirement_utils.py"))
vmod = importlib.util.module_from_spec(vspec)
vspec.loader.exec_module(vmod)

BASE_REQUIREMENTS = getattr(vmod, 'BASE_REQUIREMENTS')
EXTRAS_REQUIRE = getattr(vmod, 'EXTRAS_REQUIRE')

# If the environment has a build number set...
if os.getenv('buildnum') is not None:
    # ...append it to the version.
    version = "{version}.{buildnum}".format(
        version=version,
        buildnum=os.getenv('buildnum')
    )

setup(
    name='zenml',
    description="ZenML: Write production-ready ML code.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(
        exclude=["*.tests",
                 "*.tests.*",
                 "tests.*",
                 "tests",
                 "docs"]),
    version=version,
    install_requires=BASE_REQUIREMENTS,
    extras_require=EXTRAS_REQUIRE,
    entry_points="""
    [console_scripts]
    zenml=zenml.cli.cli:cli
    """,
    python_requires=">=3.6, <3.9.0",
    license='Apache License 2.0',  # noqa
    author='maiot GmbH',
    author_email='support@maiot.io',
    url='https://zenml.io/',
    keywords=[
        "deep", "learning", "production", "machine", "pipeline", "mlops",
        "kubernetes", "machine-learning", "deep-learning"
    ],
    # See https://PyPI.python.org/PyPI?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for.
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',

        # Pick your license.  (It should match "license" above.)

        'License :: OSI Approved :: Apache Software License',  # noqa
        # noqa
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
    ],
    # add .html file for statistics
    data_files=[('zenml/utils/post_training/stats.html',
                 ['zenml/utils/post_training/stats.html'])],
    include_package_data=True
)



import tensorflow as tf
import os
import json
import argparse
from shutil import copyfile

parser = argparse.ArgumentParser(
    description="Process images and create metadata JSON file.")

parser.add_argument("input_dir", type=str,
                    default="gs://zenml_quickstart/cycle_gan_mini",
                    help="directory containing the images.")

parser.add_argument("output_dir", type=str, default=".",
                    help="output_directory for the processed images.")

parser.add_argument("--max_imgs", type=int, default=1 << 30,
                    help="Maximum number of images to process.")

monet_dir = "monet_jpg"  # change this to your monet directory
real_dir = "photo_jpg"  # change this to your photo directory
image_ext = ".jpg"


def prepare_gan_image_dir(input_dir, output_dir, max_imgs):

    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    monet_imgs = tf.io.gfile.glob(
        os.path.join(input_dir, monet_dir, "*" + image_ext))

    real_imgs = tf.io.gfile.glob(
        os.path.join(input_dir, real_dir, "*" + image_ext))

    label_dict = {}

    for img_set in [monet_imgs, real_imgs]:
        for i, img in enumerate(img_set):
            if i == max_imgs:
                break
            with open(img, "rb") as img_file:
                monet_img_raw = img_file.read()
            monet_img = tf.io.decode_image(monet_img_raw).numpy()
            metadata = dict(zip(["height", "width", "num_channels"],
                                monet_img.shape))
            file_name = os.path.basename(img)
            copyfile(img, os.path.join(output_dir, file_name))

            label = "monet" if monet_dir in img else "real"
            label_dict[file_name] = {"label": label, "metadata": metadata}

    json_outpath = os.path.join(output_dir, "labels.json")
    with open(json_outpath, "w") as fp:
        json.dump(label_dict, fp)


if __name__ == "__main__":
    args = parser.parse_args()

    prepare_gan_image_dir(input_dir=args.input_dir,
                          output_dir=args.output_dir,
                          max_imgs=args.max_imgs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Any, Dict, Optional, Text, Union, List

from tfx import types
from tfx.dsl.components.base import base_component
from tfx.dsl.components.base import executor_spec
from tfx.orchestration import data_types
from tfx.types.component_spec import ChannelParameter
from tfx.types.component_spec import ComponentSpec
from tfx.types.component_spec import ExecutionParameter
from tfx.types.standard_artifacts import Examples, TransformGraph, Schema, \
    Model, ModelRun, HyperParameters
from tfx.utils import json_utils

from zenml.components.trainer import constants
from zenml.components.trainer.executor import ZenMLTrainerExecutor


class ZenMLTrainerSpec(ComponentSpec):
    PARAMETERS = {
        constants.RUN_FN: ExecutionParameter(type=(str, Text), optional=True),
        constants.CUSTOM_CONFIG: ExecutionParameter(type=(str, Text),
                                                    optional=True),
    }
    INPUTS = {
        constants.EXAMPLES: ChannelParameter(type=Examples),
        constants.SCHEMA: ChannelParameter(type=Schema,
                                           optional=True),
        constants.BASE_MODEL: ChannelParameter(type=Model,
                                               optional=True),
        constants.TRANSFORM_GRAPH: ChannelParameter(type=TransformGraph,
                                                    optional=True),
        constants.HYPERPARAMETERS: ChannelParameter(type=HyperParameters,
                                                    optional=True),
    }
    OUTPUTS = {
        constants.MODEL: ChannelParameter(type=Model),
        constants.MODEL_RUN: ChannelParameter(type=ModelRun),
        constants.TEST_RESULTS: ChannelParameter(type=Examples)
    }


class Trainer(base_component.BaseComponent):
    """
    A slightly adjusted version of the TFX Trainer Component. It features an
    additional output artifact to save the test results in.
    """
    SPEC_CLASS = ZenMLTrainerSpec
    EXECUTOR_SPEC = executor_spec.ExecutorClassSpec(ZenMLTrainerExecutor)

    def __init__(
            self,
            examples: types.Channel = None,
            transformed_examples: Optional[types.Channel] = None,
            transform_graph: Optional[types.Channel] = None,
            schema: Optional[types.Channel] = None,
            base_model: Optional[types.Channel] = None,
            hyperparameters: Optional[types.Channel] = None,
            run_fn: Optional[Union[Text, data_types.RuntimeParameter]] = None,
            custom_config: Optional[Dict[Text, Any]] = None,
            custom_executor_spec: Optional[executor_spec.ExecutorSpec] = None,
            output: Optional[types.Channel] = None,
            model_run: Optional[types.Channel] = None,
            test_results: Optional[types.Channel] = None,
            instance_name: Optional[Text] = None):

        if bool(examples) == bool(transformed_examples):
            raise ValueError("Exactly one of 'example' or "
                             "'transformed_example' must be supplied.")

        if transformed_examples and not transform_graph:
            raise ValueError("If 'transformed_examples' is supplied, "
                             "'transform_graph' must be supplied too.")

        examples = examples or transformed_examples
        output = output or types.Channel(type=Model)
        model_run = model_run or types.Channel(type=ModelRun)
        test_results = test_results or types.Channel(type=Examples)

        spec = ZenMLTrainerSpec(examples=examples,
                                transform_graph=transform_graph,
                                schema=schema,
                                base_model=base_model,
                                hyperparameters=hyperparameters,
                                run_fn=run_fn,
                                custom_config=json_utils.dumps(custom_config),
                                model=output,
                                model_run=model_run,
                                test_results=test_results)

        super(Trainer, self).__init__(spec=spec,
                                      custom_executor_spec=custom_executor_spec,
                                      instance_name=instance_name)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the Spark Processing Backend"""

from zenml.backends.processing import ProcessingBaseBackend


class ProcessingSparkBackend(ProcessingBaseBackend):
    """
    Use this to run pipelines on Apache Spark.

    A dedicated processing backend can be used to efficiently process large
    amounts of incoming data in parallel, potentially distributed across
    multiple machines. This can happen on base processing backends as well
    as cloud-based variants like Google Cloud Dataflow. More powerful machines
    with higher core counts and clock speeds can be leveraged to increase
    processing throughput significantly.

    This backend is not implemented yet.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        raise NotImplementedError('Its coming soon!')



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import List, Text

import tensorflow as tf

from trainer.gan_functions import Generator, Discriminator, \
    CycleGan, TensorBoardImage
from trainer.gan_functions import discriminator_loss, \
    identity_loss, generator_loss, calc_cycle_loss
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.utils.naming_utils import check_if_transformed_feature, \
    check_if_transformed_label


class CycleGANTrainer(TFFeedForwardTrainer):
    def __init__(self, batch_size=1, epochs=25, **kwargs):
        self.batch_size = batch_size
        self.epochs = epochs
        super(CycleGANTrainer, self).__init__(batch_size=batch_size,
                                              epochs=epochs,
                                              **kwargs)

    def model_fn(self,
                 train_dataset: tf.data.Dataset,
                 eval_dataset: tf.data.Dataset):
        dataset = tf.data.Dataset.zip((train_dataset, eval_dataset))

        monet_generator_optimizer = tf.keras.optimizers.Adam(2e-4, beta_1=0.5)
        photo_generator_optimizer = tf.keras.optimizers.Adam(2e-4, beta_1=0.5)

        monet_discriminator_optimizer = tf.keras.optimizers.Adam(2e-4,
                                                                 beta_1=0.5)
        photo_discriminator_optimizer = tf.keras.optimizers.Adam(2e-4,
                                                                 beta_1=0.5)

        # transforms photos to Monet paintings
        monet_generator = Generator()
        # transforms Monet paintings to be more like photos
        photo_generator = Generator()

        # differentiates real Monet paintings and generated Monet paintings
        monet_discriminator = Discriminator()
        # differentiates real photos and generated photos
        photo_discriminator = Discriminator()

        cycle_gan_model = CycleGan(
            monet_generator, photo_generator, monet_discriminator,
            photo_discriminator
        )

        cycle_gan_model.compile(
            m_gen_optimizer=monet_generator_optimizer,
            p_gen_optimizer=photo_generator_optimizer,
            m_disc_optimizer=monet_discriminator_optimizer,
            p_disc_optimizer=photo_discriminator_optimizer,
            gen_loss_fn=generator_loss,
            disc_loss_fn=discriminator_loss,
            cycle_loss_fn=calc_cycle_loss,
            identity_loss_fn=identity_loss
        )

        cycle_gan_model.fit(
            dataset,
            batch_size=self.batch_size,
            epochs=self.epochs,
            callbacks=[
                tf.keras.callbacks.TensorBoard(
                    log_dir=os.path.join(self.log_dir, 'train')),
                TensorBoardImage(test_data=eval_dataset,
                                 log_dir=os.path.join(self.log_dir, 'img'))]
        )

        return cycle_gan_model

    def input_fn(self,
                 file_pattern: List[Text]):
        xf_feature_spec = {x: self.schema[x]
                           for x in self.schema
                           if check_if_transformed_feature(x)
                           or check_if_transformed_label(x)}

        return tf.data.experimental.make_batched_features_dataset(
            file_pattern=file_pattern,
            batch_size=self.batch_size,
            shuffle=False,
            num_epochs=1,
            features=xf_feature_spec,
            reader=self._gzip_reader_fn)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

"""
This collection of utility functions is aiming to help the users follow a
certain pre-defined naming paradigm.

However, due to the highly dynamic nature of ML tasks, users might need to
implement their own naming scheme depending on their tasks.
"""

from typing import Text

TRANSFORM_SUFFIX = '_xf'
LABEL_SUFFIX = '_xl'
OUTPUT_SUFFIX = '_o'


def transformed_feature_name(name: Text):
    return name + TRANSFORM_SUFFIX


def transformed_label_name(name: Text):
    return name + LABEL_SUFFIX


def output_name(name: Text):
    return name + OUTPUT_SUFFIX


def check_if_transformed_feature(name: Text):
    return name.endswith(TRANSFORM_SUFFIX)


def check_if_transformed_label(name: Text):
    return name.endswith(LABEL_SUFFIX)


def check_if_output_name(name: Text):
    return name.endswith(OUTPUT_SUFFIX)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
import random
from pathlib import Path
from typing import Text

import pytest

import zenml
from zenml.datasources import BaseDatasource
from zenml.pipelines import BasePipeline
from zenml.repo import Repository
from zenml.utils import yaml_utils
from zenml.version import __version__

# Nicholas a way to get to the root
ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")


def test_repo_double_init():
    # explicitly constructing another repository should fail
    with pytest.raises(Exception):
        _ = Repository()


def test_get_datasources(repo):
    ds_list = repo.get_datasources()

    assert "my_csv_datasource" in [x.name for x in ds_list]


def test_get_datasource_by_name(repo):
    assert repo.get_datasource_by_name("my_csv_datasource")

    fake_ds = repo.get_datasource_by_name("ds_123")

    assert fake_ds is None


def test_get_datasource_names(repo):
    # TODO [LOW]: Automatically expand when new datasource tests are added!
    test_ds_names = ["my_csv_datasource", "image_ds_local", "image_ds_gcp",
                     "json_ds"]

    ds_names = repo.get_datasource_names()

    assert set(test_ds_names) <= set(ds_names)


def test_get_pipeline_file_paths(repo, monkeypatch):
    mock_paths = ["pipeline_1.yaml", "pipeline_2.yaml", "awjfof.txt"]

    def mock_list_dir(dir_path: Text, only_file_names: bool = False):
        # add a corrupted file into the pipelines
        return mock_paths

    monkeypatch.setattr("zenml.utils.path_utils.list_dir",
                        mock_list_dir)

    paths = repo.get_pipeline_file_paths(only_file_names=True)

    assert paths == mock_paths[:-1]


def test_get_pipeline_names(repo):
    # TODO: This has to be made dynamic once more pipelines come
    real_p_names = sorted(["csvtest{0}".format(i) for i in range(1, 6)])

    found_p_names = sorted(repo.get_pipeline_names())

    assert set(real_p_names) <= set(found_p_names)


def test_get_pipelines(repo):
    p_names = sorted(repo.get_pipeline_names())

    pipelines = repo.get_pipelines()

    pipelines = sorted(pipelines, key=lambda p: p.name)

    assert all(p.name == name for p, name in zip(pipelines, p_names))


def test_get_pipelines_by_datasource(repo):
    # asserted in an earlier test
    ds = repo.get_datasource_by_name("my_csv_datasource")

    ds2 = BaseDatasource(name="ds_12254757")

    pipelines = repo.get_pipelines_by_datasource(ds)

    pipelines_2 = repo.get_pipelines_by_datasource(ds2)

    assert len(pipelines) > 0

    assert not pipelines_2


def test_get_pipelines_by_type(repo):
    pipelines = repo.get_pipelines_by_type(type_filter=["training"])

    pipelines_2 = repo.get_pipelines_by_type(type_filter=["base"])

    assert len(pipelines) == 5

    assert not pipelines_2


def test_get_pipeline_by_name(repo, equal_pipelines):
    p_names = repo.get_pipeline_names()

    random_name = random.choice(p_names)
    cfg_list = [y for y in repo.get_pipeline_file_paths()
                if random_name in y]

    cfg = yaml_utils.read_yaml(cfg_list[0])

    p1 = repo.get_pipeline_by_name(random_name)

    p2 = BasePipeline.from_config(cfg)

    assert equal_pipelines(p1, p2, loaded=True)


def test_get_step_versions(repo):
    step_versions = repo.get_step_versions()

    # TODO: Make this less hardcoded
    steps_used = [
        "zenml.steps.preprocesser.standard_preprocesser.standard_preprocesser.StandardPreprocesser",
        "zenml.steps.split.categorical_domain_split_step.CategoricalDomainSplit",
        "zenml.steps.trainer.tensorflow_trainers.tf_ff_trainer.FeedForwardTrainer"
    ]

    current_version = "zenml_" + str(__version__)

    assert set(steps_used) >= set(step_versions.keys())
    assert all(current_version in s for s in step_versions.values())


def test_get_step_by_version(repo):
    # TODO: Make this less hardcoded
    steps_used = [
        "zenml.steps.preprocesser.standard_preprocesser.standard_preprocesser.StandardPreprocesser",
        "zenml.steps.split.categorical_domain_split_step.CategoricalDomainSplit",
        "zenml.steps.trainer.tensorflow_trainers.tf_ff_trainer.FeedForwardTrainer"
    ]

    random_step = random.choice(steps_used)

    current_version = "zenml_" + str(__version__)

    bogus_version = "asdfghjklöä"

    assert repo.get_step_by_version(random_step, current_version)
    assert repo.get_step_by_version(random_step, bogus_version) is None


def test_get_step_versions_by_type(repo):
    # TODO: Make this less hardcoded
    steps_used = [
        "zenml.steps.preprocesser.standard_preprocesser.standard_preprocesser.StandardPreprocesser",
        "zenml.steps.split.categorical_domain_split_step.CategoricalDomainSplit",
        "zenml.steps.trainer.tensorflow_trainers.tf_ff_trainer.FeedForwardTrainer"
    ]

    random_step = random.choice(steps_used)

    current_version = "zenml_" + str(__version__)

    bogus_step = "asdfghjklöä"

    step_versions = repo.get_step_versions_by_type(random_step)

    assert step_versions == {current_version}

    assert repo.get_step_versions_by_type(bogus_step) is None



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of a the base Backend"""

from typing import Dict

from zenml.logger import get_logger
from zenml.standards.standard_keys import BackendKeys
from zenml.utils import source_utils
from zenml.utils.print_utils import to_pretty_string, PrintStyles

logger = get_logger(__name__)


class BaseBackend:
    """
    Base class for all ZenML backends.

    Every ZenML pipeline runs in backends that defines where and how the
    pipeline runs. Override this base class to define your own custom Pipeline
    backend.

    There are three types of backends available in ZenML: Orchestration
    backends, processing backends and training backends. Each of them serve
    different purposes in different stages of the pipeline.
    An orchestration backend is useful for scheduling and executing the
    different pipeline components.

    A dedicated processing backend can be used to efficiently process large
    amounts of incoming data in parallel, potentially distributed across
    multiple machines. This can happen on local processing backends as well
    as cloud-based variants like Google Cloud Dataflow. More powerful machines
    with higher core counts and clock speeds can be leveraged to increase
    processing throughput significantly.

    A training backend can be used to efficiently train a machine learning
    model on large amounts of data. Since most common machine learning models
    leverage mainly linear algebra operations under the hood, they can
    potentially benefit a lot from dedicated training hardware like Graphics
    Processing Units (GPUs) or application-specific integrated circuits
    (ASICs). Again, local training backends or cloud-based training backends
    like Google Cloud AI Platform (GCAIP) with or without GPU/ASIC support
    can be used.
    """
    BACKEND_TYPE = None

    def __init__(self, **kwargs):
        if len(kwargs):
            self._kwargs = kwargs
        else:
            self._kwargs = {}

        self._source = source_utils.resolve_class(self.__class__)
        self._immutable = False

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    def to_config(self):
        """Converts Backend to ZenML config block."""
        return {
            BackendKeys.TYPE: self.BACKEND_TYPE,
            BackendKeys.ARGS: self._kwargs,  # everything in init
            BackendKeys.SOURCE: self._source,
        }

    @classmethod
    def from_config(cls, config: Dict):
        """
        Convert from ZenML config dict to ZenML Backend object.

        Args:
            config: a ZenML config in dict-form (probably loaded from YAML)
        """
        backend_class = source_utils.load_source_path_class(
            config[BackendKeys.SOURCE])
        backend_args = config[BackendKeys.ARGS]
        obj = backend_class(**backend_args)
        obj._immutable = True
        return obj
